"""
.. note::
	If you found yourself here not being a dev, you should probably :mod:`skip this section <blueprints.world>` from or 
	look up the correspinging docs page of the Thing you were looking after for which this module only 
	implements its corresponding mirror types. A link to the docs page is to be found in the Types doc 
	string.

This module is used to model the inheritance hierarchy and to define the different attribute types.
Since blueprint specifies many values of different data types, the user will easily find themselves 
setting attributes to the wrong types. But for most attribute setters this would not be caught 
resulting in a downstream exception during the XML build. Those exceptions are not easily associated 
to a line in the users code, which violated a type constraint.

To prevent this, blueprints is type safe in attributes (which is most of the user accessable interface). 
The module typechecker is used to @restrict methods to only take arguments which satisfy the type 
of the method signature. In python those type hints are not binding, @restrict however will raise 
ValueError if a type hint is violated.

To make all types available during class definition for @restrict, the types module defines a mirror 
type for all Things. Those mirror types are to be used in the type hints for @restrict instead of the 
proper Thing class.
Every Thing class in blueprints must inherit (as its first base class) from a Type class in this module.
Type classes must mimic the inheritance of its corresponding class. This is used for type hints as
well as to ensure that the inheritance structure is preserved, when ``isinstance`` or ``issubclass``
is used.

Additionally the mirror types are used to specify class properties like data type associations for 
attributes used to translate back and forth between mujoco xml and blueprints.
"""


#from abc import ABC, ABCMeta
import numpy as np



class PathType(object):

	"""
	This abstract base class is a used to represent the Tendon class in type hints. For a detailed
	description see :class:`blueprints.tendon.Path`.
	"""

class FunctionHandleType(object):

	"""
	This abstract base class is a used to represent the FunctionHandle class in type hints. For a detailed
	description see :class:`blueprints.utils.view.FunctionHandle`.
	"""


class ViewType(object):

	"""
	This abstract base class is a used to represent the View class in type hints. For a detailed
	description see :class:`blueprints.utils.view.View`.
	"""


class AllViewType(ViewType):

	"""
	This abstract base class is a used to represent the AllView class in type hints. For a detailed
	description see :class:`blueprints.utils.view.AllView`.
	"""


class LatticeType(object):

	"""
	This abstract base class is a used to represent the Lattice class in type hints. For a detailed
	description see :class:`blueprints.utils.lattice.Lattice`.
	"""


class LatticeViewType(object):

	"""
	This abstract base class is a used to represent the LatticeView class in type hints. For a detailed
	description see :class:`blueprints.utils.view.LatticeView`.
	"""


class ColorType(object):

	"""
	This abstract base class is a used to represent the View class in type hints. For a detailed
	description see :class:`blueprints.utils.view.View`.
	"""

	_COLORS = {'red':         [1.0,  0.0,   0.0],
		   'orange':      [1.0,  0.5,   0.0],
		   'yellow':      [0.8,  0.8,   0.0],
		   'green':       [0.0,  1.0,   0.0],
		   'teal':        [0.0,  1.0,   1.0],
		   'blue':        [0.0,  0.0,   1.0],
		   'purple':      [1.0,  0.0,   1.0],
		   'black':       [0.0,  0.0,   0.0],
		   'grey':        [0.5,  0.5,   0.5],
		   'white':       [1.0,  1.0,   1.0], 
		   'dark_red':    [0.5,  0.0,   0.0],
		   'dark_orange': [0.5,  0.25,  0.0],
		   'dark_yellow': [0.4,  0.4,   0.0],
		   'dark_green':  [0.0,  0.5,   0.0],
		   'dark_teal':   [0.0,  0.5,   0.5],
		   'dark_blue':   [0.0,  0.0,   0.5],
		   'dark_purple': [0.5,  0.0,   0.5],
		   'dark_grey':   [0.25, 0.25,  0.25]}


class ThingType(object):

	"""
	This abstract base class is a used to represent the Thing class in type hints. For a detailed
	description see :class:`blueprints.thing.base.BaseThing`.
	"""

	_NEW_NO_COPY_ATTR   = {'parent'}
	_NEW_DEFAULT_VALS   = dict()
	_NEW_DERIVED_ATTR   = dict()
	_NEW_MUJOCO_ATTR    = {'name':   str}



ThingType._NEW_BLUEPRINT_ATTR = {'parent': ThingType}



class NodeThingType(ThingType):

	"""
	This abstract base class is a used to represent the NodeThing class in type hints. For a detailed
	description see :class:`blueprints.thing.node.NodeThing`.
	"""

	_CYCLE_REF = dict()


class MoveableThingType(ThingType):

	"""
	This abstract base class is a used to represent the MoveableThing class in type hints. For a detailed
	description see :class:`blueprints.thing.moveable.MoveableThing`.
	"""
	_NEW_STEP_CACHE     = {'pos', 
			       'rotation_matrix', 
			       'global_rotation_matrix', 
			       'euler', 
			       'global_pos', 
			       'vel', 
			       'angular_vel'}
	_NEW_DEFAULT_VALS   = {'pos':   np.array([0., 0., 0.], dtype=np.float32),
			       'euler': np.array([0., 0., 0.], dtype=np.float32),
			       'alpha': 0.,
			       'beta':  0.,
			       'gamma': 0.}
	_NEW_BLUEPRINT_ATTR = {'pos':   np.ndarray,
			       'global_pos': np.ndarray, 
			       'alpha': float,
			       'beta':  float,
			       'gamma': float}
	_NEW_MUJOCO_ATTR    = {'euler':    np.ndarray,
			       'pos':      np.ndarray}


class ColoredThingType(ThingType):

	"""
	This abstract base class is a used to represent the ColoredThing class in type hints. For a detailed
	description see :class:`blueprints.thing.colored.ColoredThing`.
	"""

	_NEW_DEFAULT_VALS   = {'rgba':    np.array([.5, .5, .5, 1.], dtype=np.float32),
			       'opacity': 1.}
	_NEW_BLUEPRINT_ATTR = {'color': ColorType}
	_NEW_MUJOCO_ATTR    = {'rgba':  np.ndarray}


class CyclicalThingType(ThingType):

	"""
	This abstract base class is a used to represent the CyclicalThing class in type hints. For a detailed
	description see :class:`blueprints.thing.cyclical.CyclicalThing`.
	"""


class UniqueThingType(ThingType):

	"""
	This abstract base class is a used to represent the UniqueThing class in type hints. For a detailed
	description see :class:`blueprints.thing.unique.UniqueThing`.
	"""


class FocalThingType(ThingType):

	"""
	This abstract base class is a used to represent the FocalThing class in type hints. For a detailed
	description see :class:`blueprints.thing.unique.FocalThing`.
	"""


class CacheType(UniqueThingType):

	"""
	This abstract base class is a used to represent the Cache class in type hints. For a detailed
	description see :class:`blueprints.cache.BaseCache`.
	"""

	_REFERENCE_NAME = 'cache'


class MeshCacheType(CacheType):

	"""
	This abstract base class is a used to represent the MeshCache class in type hints. For a detailed
	description see :class:`blueprints.cache.MeshCache`.
	"""

	_NEW_BLUEPRINT_ATTR = {'vertecies':      list,
			       'faces':          list,
			       'texcoords':      list,
			       'texcoords_idx':  dict,
			       'normals_idx':    dict,
			       'face_normals':   list,
			       'vertex_normals': list,
			       'filename':       str,
			       'centered':       bool}
	_NEW_MUJOCO_ATTR    = {'vertex':   np.ndarray,
			       'face':     np.ndarray}
	_ASSET_OBJ          =  'mesh'


class HFieldCacheType(CacheType):

	"""
	This abstract base class is a used to represent the MeshCache class in type hints. For a detailed
	description see :class:`blueprints.cache.MeshCache`.
	"""

	_NEW_BLUEPRINT_ATTR = {'terrain':   list,
			       'filename':  str}
	_NEW_MUJOCO_ATTR    = {'elevation': np.ndarray, 
			       'nrow':      int, 
			       'ncol':      int}
	_ASSET_OBJ          =  'hfield'


class AssetType(UniqueThingType):

	"""
	This abstract base class is a used to represent the Asset class in type hints. For a detailed
	description see :class:`blueprints.assets.BaseAsset`.
	"""

	_REFERENCE_NAME = 'asset'



class TextureAssetType(AssetType):

	"""
	This abstract base class is a used to represent the TextureAsset class in type hints. For a detailed
	description see :class:`blueprints.assets.TextureAsset`.
	"""

	_NEW_MUJOCO_ATTR    = {'type':         str, 
			       #'colorspace':  str, 
			       'content_type': str, 
			       'file':         str, 
			       'gridsize':     np.ndarray, 
			       'gridlayout':   str, 
			       'fileright':    str, 
			       'fileleft':     str, 
			       'fileup':       str, 
			       'filedown':     str, 
			       'filefront':    str, 
			       'fileback':     str, 
			       'builtin':      str, 
			       'rgb1':         np.ndarray, 
			       'rgb2':         np.ndarray, 
			       'mark':         str, 
			       'markrgb':      np.ndarray, 
			       'random':       float, 
			       'width':        int, 
			       'height':       int, 
			       'hflip':        bool, 
			       'vflip':        bool, 
			       'nchannel':     int}
	_NEW_BLUEPRINT_ATTR = {#'color_space': str, 
			       'content':      str, 
			       'file':         str, 
			       'grid_size':    list, 
			       'grid_layout':  str, 
			       'fileright':    str, 
			       'fileleft':     str, 
			       'fileup':       str, 
			       'filedown':     str, 
			       'filefront':    str, 
			       'fileback':     str, 
			       'builtin':      str, 
			       'color_1':      ColorType, 
			       'color_2':      ColorType, 
			       'mark':         str, 
			       'color_mark':   ColorType, 
			       'random':       float, 
			       'width':        int, 
			       'height':       int, 
			       'h_flip':       bool, 
			       'v_flip':       bool, 
			       'n_channel':    int}
	_NEW_DEFAULT_VALS   = {'gridsize':     np.array([1, 1], dtype=np.int8), 
			       'gridlayout':   '…………',  
			       'rgb1':         np.array([0.8, 0.8, 0.8], dtype=np.float32), 
			       'rgb2':         np.array([0.5, 0.5, 0.5], dtype=np.float32), 
			       'markrgb':      np.array([0.0, 0.0, 0.0], dtype=np.float32), 
			       'random':       0.01, 
			       'width':        0, 
			       'height':       0, 
			       'hflip':        False, 
			       'vflip':        False, 
			       'nchannel':     3}
	_REFERENCE_NAME = 'asset'



class TextureType(ThingType):

	"""
	This abstract base class is a used to represent the Texture class in type hints. For a detailed
	description see :class:`blueprints.texture.BaseTexture`.
	"""
	_NEW_NO_COPY_ATTR      = {'asset'}
	_NEW_SINGLE_CHILD_ATTR = {'asset':       AssetType}
	_NEW_BLUEPRINT_ATTR    = {'asset':       TextureAssetType}



class PlaneTextureType(TextureType):

	"""
	This abstract base class is a used to represent the Texture class in type hints. For a detailed
	description see :class:`blueprints.texture.Plane`.
	"""

	_TYPE = '2d'



class BoxTextureType(TextureType):

	"""
	This abstract base class is a used to represent the Texture class in type hints. For a detailed
	description see :class:`blueprints.texture.Box`.
	"""

	_TYPE = 'cube'



class SkyboxTextureType(BoxTextureType):

	"""
	This abstract base class is a used to represent the Texture class in type hints. For a detailed
	description see :class:`blueprints.texture.Skybox`.
	"""

	_TYPE = 'skybox'



class MaterialAssetType(AssetType, ColoredThingType):

	"""
	This abstract base class is a used to represent the MaterialAsset class in type hints. For a detailed
	description see :class:`blueprints.assets.MaterialAsset`.
	"""

	_NEW_MUJOCO_ATTR    = {'texrepeat':   np.ndarray, 
			       'texuniform':  bool, 
			       'emission':    float, 
			       'specular':    float, 
			       'shininess':   float, 
			       'reflectance': float, 
			       'metallic':    float, 
			       'roughness':   float}
	_NEW_BLUEPRINT_ATTR = {'texture':     TextureType, 
			       'texrepeat':   np.ndarray, 
			       'texuniform':  bool, 
			       'emission':    float, 
			       'specular':    float, 
			       'shininess':   float, 
			       'reflectance': float, 
			       'metallic':    float, 
			       'roughness':   float}
	_NEW_DEFAULT_VALS   = {'texrepeat':   np.array([1., 1.], dtype=np.float32), 
			       'texuniform':  False, 
			       'emission':    0.0, 
			       'specular':    0.5, 
			       'shininess':   0.5, 
			       'reflectance': 0.0, 
			       'metallic':   -1.0, 
			       'roughness':  -1.0}



class MaterialType(NodeThingType):

	"""
	This abstract base class is a used to represent the Material class in type hints. For a detailed
	description see :class:`blueprints.Material`.
	"""

	_NEW_NO_COPY_ATTR      = {'asset'}
	_NEW_SINGLE_CHILD_ATTR = {'asset':       AssetType}
	_NEW_BLUEPRINT_ATTR    = {'texture':     TextureAssetType, 
				  'texrepeat':   np.ndarray, 
				  'texuniform':  bool, 
				  'emission':    float, 
				  'specular':    float, 
				  'shininess':   float, 
				  'reflectance': float, 
				  'metallic':    float, 
				  'roughness':   float}
	



class MeshAssetType(MoveableThingType, AssetType):

	"""
	This abstract base class is a used to represent the MeshAsset class in type hints. For a detailed
	description see :class:`blueprints.assets.MeshAsset`.
	"""

	_NEW_DEFAULT_VALS   = {'scale':    np.array([1., 1., 1.], dtype=np.float32),
			       'refpos':   np.array([0., 0., 0.], dtype=np.float32)}
	_NEW_DERIVED_ATTR   = {'vertex':   np.ndarray}
	_NEW_BLUEPRINT_ATTR = {'scale':    np.ndarray,
			       'filename': str,
			       'cache':    CacheType,
			       'xml_data': bool}
	_NEW_MUJOCO_ATTR    = {'refpos':   np.ndarray,
			       'scale':    np.ndarray}
	_DEL_MUJOCO_ATTR    = {'euler',
			       'pos'}
	_ASSET_OBJ          =  'mesh'


class HFieldAssetType(MoveableThingType, AssetType):

	"""
	This abstract base class is a used to represent the MeshAsset class in type hints. For a detailed
	description see :class:`blueprints.assets.MeshAsset`.
	"""

	_NEW_DEFAULT_VALS   = {'pos':           np.array([0., 0., 0.], dtype=np.float32), 
			       'x_length':      1., 
			       'y_length':      1., 
			       'z_length':      1., 
			       'height_offset': 1.}
	_NEW_DERIVED_ATTR   = {'elevation':     np.ndarray}
	_NEW_MUJOCO_ATTR    = {'size':          np.ndarray}
	_DEL_MUJOCO_ATTR    = {'pos'}
	_NEW_BLUEPRINT_ATTR = {'filename':      str,
			       'cache':         CacheType, 
			       'xml_data':      bool, 
			       'x_length':      float, 
			       'y_length':      float, 
			       'z_length':      float}
	_ASSET_OBJ          =  'hfield'


class WorldType(NodeThingType):

	"""
	This abstract base class is a used to represent the World class in type hints. For a detailed
	description see :class:`blueprints.world.World`.
	"""

	_MUJOCO_OBJ = 'world'


class BodyType(NodeThingType):

	"""
	This abstract base class is a used to represent the Body class in type hints. For a detailed
	description see :class:`blueprints.body.Body`.
	"""

	_NEW_MUJOCO_ATTR = {'euler': np.ndarray,
		            'pos':   np.ndarray}
	_MUJOCO_OBJ      =  'body'
	_MUJOCO_DATA     =  'body'


class PlaceholderType(MoveableThingType):

	"""
	This abstract base class is a used to represent the Placeholder class in type hints. For a detailed
	description see :class:`blueprints.placeholder.Placeholder`.
	"""


class TubeType(ThingType):

	"""
	This abstract base class is a used to represent the Tube class in type hints. For a detailed
	description see :class:`blueprints.tube.BaseTube`.
	"""


class GeomType(MoveableThingType, ColoredThingType, NodeThingType):

	"""
	This abstract base class is a used to represent the Geom class in type hints. For a detailed
	description see :class:`blueprints.geoms.BaseGeom`.
	"""

	_NEW_SINGLE_CHILD_ATTR = {'material':           MaterialType}
	_NEW_DEFAULT_VALS      = {'friction':           np.array([1.0, 0.005, 0.0001], dtype=np.float32),
				  'margin':             0.,
				  'gap':                0.,
				  'shellinertia':       False,
				  'mass':               None,
				  'density':            1000.}
	_NEW_DERIVED_ATTR      = {'type':               str}
	_NEW_BLUEPRINT_ATTR    = {'mass':               float,
				  'density':            float,
				  'shellinertia':       bool,
				  'margin':             float,
				  'gap':                float,
				  'sliding_friction':   float,
				  'torsional_friction': float,
				  'rolling_friction':   float, 
				  'material':           MaterialType}
	_NEW_MUJOCO_ATTR       = {'size':               np.ndarray,
				  'mass':               float,
				  'density':            float,
				  'shellinertia':       bool,
				  'margin':             float,
				  'gap':                float,
				  'friction':           np.ndarray,
				  'type':               str}
	_MUJOCO_OBJ             = 'geom'
	_MUJOCO_DATA            = 'geom'


class CapsuleGeomType(GeomType, TubeType):

	"""
	This abstract base class is a used to represent the CapsuleGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Capsule`.
	"""

	_NEW_BLUEPRINT_ATTR = {'radius': float,
			       'length': float}


class CylinderGeomType(GeomType, TubeType):

	"""
	This abstract base class is a used to represent the CylinderGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Cylinder`.
	"""

	_NEW_BLUEPRINT_ATTR = {'radius': float,
			       'length': float}


class BoxGeomType(GeomType, TubeType):

	"""
	This abstract base class is a used to represent the BoxGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Box`.
	"""

	_NEW_BLUEPRINT_ATTR = {'x_length': float,
			       'y_length': float,
			       'z_length': float}


class PlaneGeomType(GeomType):

	"""
	This abstract base class is a used to represent the PlaneGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Plane`.
	"""

	_NEW_BLUEPRINT_ATTR = {'x_length': float,
			       'y_length': float,
			       'spacing':  float}


class SphereGeomType(GeomType):

	"""
	This abstract base class is a used to represent the SphereGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Sphere`.
	"""

	_NEW_BLUEPRINT_ATTR = {'radius': float}


class EllipsoidGeomType(GeomType):

	"""
	This abstract base class is a used to represent the EllipsoidGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Ellipsoid`.
	"""

	_NEW_BLUEPRINT_ATTR = {'x_radius': float,
			       'y_radius': float,
			       'z_radius': float}


class MeshGeomType(GeomType):

	"""
	This abstract base class is a used to represent the MeshGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Mesh`.
	"""

	_NEW_SINGLE_CHILD_ATTR = {'asset':    AssetType, 
				  'material': MaterialType}
	_NEW_NO_COPY_ATTR      = {'asset'}
	_NEW_BLUEPRINT_ATTR    = {'asset': AssetType}
	#_DEL_BLUEPRINT_ATTR = {'pos'}
	_DEL_MUJOCO_ATTR       = {'pos',
				  'size'}
	_MUJOCO_DATA           =  'mesh'


class HFieldGeomType(GeomType):

	"""
	This abstract base class is a used to represent the HfieldGeom class in type hints. For a detailed
	description see :class:`blueprints.geoms.Hfield`.
	"""

	_NEW_SINGLE_CHILD_ATTR = {'asset': AssetType}
	_NEW_BLUEPRINT_ATTR    = {'asset': AssetType}
	_DEL_MUJOCO_ATTR       = {'size'}


class SiteType(MoveableThingType, NodeThingType):

	"""
	This abstract base class is a used to represent the Site class in type hints. For a detailed
	description see :class:`blueprints.sites.BaseSite`.
	"""

	_NEW_DEFAULT_VALS   = {'size': np.array([0.005, 0.005, 0.005], dtype=np.float32)}
	_NEW_DERIVED_ATTR   = {'type': str}
	_NEW_MUJOCO_ATTR    = {'size': np.ndarray,
			       'rgba': np.ndarray,
			       'type': str}
	_MUJOCO_OBJ	    =  'site'
	_MUJOCO_DATA        =  'site'


class CapsuleSiteType(TubeType, SiteType):

	"""
	This abstract base class is a used to represent the CapsuleSite class in type hints. For a detailed
	description see :class:`blueprints.sites.Capsule`.
	"""

	_NEW_BLUEPRINT_ATTR = {'radius': float,
			       'length': float}


class CylinderSiteType(TubeType, SiteType):

	"""
	This abstract base class is a used to represent the CylinderSite class in type hints. For a detailed
	description see :class:`blueprints.sites.Cylinder`.
	"""

	_NEW_BLUEPRINT_ATTR = {'radius': float,
			       'length': float}


class BoxSiteType(TubeType, SiteType):

	"""
	This abstract base class is a used to represent the BoxSite class in type hints. For a detailed
	description see :class:`blueprints.sites.Box`.
	"""

	_NEW_BLUEPRINT_ATTR = {'x_length': float,
			       'y_length': float,
			       'z_length': float}


class SphereSiteType(SiteType):

	"""
	This abstract base class is a used to represent the SphereSite class in type hints. For a detailed
	description see :class:`blueprints.sites.Sphere`.
	"""

	_NEW_BLUEPRINT_ATTR = {'radius': float}


class EllipsoidSiteType(SiteType):

	"""
	This abstract base class is a used to represent the EllipsoidSite class in type hints. For a detailed
	description see :class:`blueprints.sites.Ellipsoid`.
	"""

	_NEW_BLUEPRINT_ATTR = {'x_radius': float,
			       'y_radius': float,
			       'z_radius': float}


class JointType(NodeThingType):

	"""
	This abstract base class is a used to represent the Joint class in type hints. For a detailed
	description see :class:`blueprints.joints.BaseJoint`.
	"""

	_NEW_DEFAULT_VALS   = {'springdamper':       np.array([0., 0.], dtype=np.float32),
			       'actuatorforcerange': np.array([0., 0.], dtype=np.float32),
			       'stiffness':          0.,
			       'springref':          0.,
			       'armature':           0.,
			       'damping':            0.}
	_NEW_DERIVED_ATTR   = {'type':               str}
	_NEW_BLUEPRINT_ATTR = {'springdamper':       np.ndarray,
			       'actuatorforcerange': np.ndarray,
			       'stiffness':          float,
			       'springref':          float,
			       'armature':           float,
			       'damping':            float,
			       #'sensors':      list,
			       #'actuators':    list,
			       }
	_NEW_MUJOCO_ATTR    = {'springdamper':       np.ndarray,
			       'actuatorforcerange': np.ndarray,
			       'stiffness':          np.ndarray,
			       'springref':          np.ndarray,
			       'armature':           np.ndarray,
			       'damping':            np.ndarray,
			       'type':               str}
	_DEL_MUJOCO_ATTR    = {'euler'}
	_DEL_BLUEPRINT_ATTR = {'euler'}
	_MUJOCO_OBJ         =  'joint'
	_MUJOCO_DATA        =  'jnt'


class HingeType(JointType):

	"""
	This abstract base class is a used to represent the Hinge class in type hints. For a detailed
	description see :class:`blueprints.joints.Hinge`.
	"""

	_NEW_DEFAULT_VALS   = {'axis':        [0., 0., 1.],
			       'range':       [0., 0.],
			       'ref':          0,
			       'frictionloss': 0.}
	_NEW_BLUEPRINT_ATTR = {'axis':         np.ndarray,
			       'range':        np.ndarray,
			       'ref':          float,
			       'frictionloss': float}
	_NEW_MUJOCO_ATTR    = {'axis':         np.ndarray,
			       'range':        np.ndarray,
			       'ref':          float,
			       'frictionloss': float}


class SlideType(JointType):

	"""
	This abstract base class is a used to represent the Slide class in type hints. For a detailed
	description see :class:`blueprints.joints.Slide`.
	"""

	_NEW_DEFAULT_VALS   = {'axis':        [0., 0., 1.],
			       'range':       [0., 0.],
			       'ref':          0,
			       'frictionloss': 0.}
	_NEW_BLUEPRINT_ATTR = {'axis':         np.ndarray,
			       'range':        np.ndarray,
			       'ref':          float,
			       'frictionloss': float}
	_NEW_MUJOCO_ATTR    = {'axis':         np.ndarray,
			       'range':        np.ndarray,
			       'ref':          float,
			       'frictionloss': float}


class BallType(JointType):

	"""
	This abstract base class is a used to represent the Ball class in type hints. For a detailed
	description see :class:`blueprints.joints.Ball`.
	"""

	_NEW_DEFAULT_VALS   = {'range':       [0., 0.],
			       'frictionloss': 0.}
	_NEW_BLUEPRINT_ATTR = {'range':        np.ndarray,
			       'frictionloss': float}
	_NEW_MUJOCO_ATTR    = {'range':        np.ndarray,
			       'frictionloss': float}


class FreeType(JointType):

	"""
	This abstract base class is a used to represent the Free class in type hints. For a detailed
	description see :class:`blueprints.joints.Free`.
	"""


class SensorType(ThingType):

	"""
	This abstract base class is a used to represent the Sensor class in type hints. For a detailed
	description see :class:`blueprints.sensors.BaseSensor`.
	"""

	_NEW_DEFAULT_VALS   = {'noise':  0.,
			       'cutoff': 0.}
	_NEW_BLUEPRINT_ATTR = {'noise':  float,
			       'cutoff': float}
	_NEW_MUJOCO_ATTR    = {'noise':  float,
			       'cutoff': float}
	_REFERENCE_TYPES    = {ThingType}
	_MUJOCO_OBJ	    =  'sensor'


class SiteSensorType(SensorType):

	"""
	This abstract base class is a used to represent the SiteSensor class in type hints. For a detailed
	description see :class:`blueprints.sensors.SiteSensor`.
	"""

	_PARENT_TYPE         =  'site'
	_NEW_DERIVED_ATTR    = {'site': str}
	_NEW_MUJOCO_ATTR     = {'site': str}


class JointSensorType(SensorType):

	"""
	This abstract base class is a used to represent the JointSensor class in type hints. For a detailed
	description see :class:`blueprints.sensors.JointSensor`.
	"""

	_PARENT_TYPE         =  'joint'
	_NEW_DERIVED_ATTR    = {'joint': str}
	_NEW_MUJOCO_ATTR     = {'joint': str}


class ActuatorSensorType(SensorType):

	"""
	This abstract base class is a used to represent the ActuatorSensor class in type hints. For a detailed
	description see :class:`blueprints.sensors.ActuatorSensor`.
	"""

	_PARENT_TYPE         =  'actuator'
	_NEW_DERIVED_ATTR    = {'actuator': str}
	_NEW_MUJOCO_ATTR     = {'actuator': str}


class TendonType(FocalThingType):

	"""
	This abstract base class is a used to represent the Tendon class in type hints. For a detailed
	description see :class:`blueprints.tendon.Tendon`.
	"""

	_FIXED_ATTR         = {'name', 
			       'limited', 
			       'range', 
			       'solreflimit', 
			       'solimplimit', 
			       'solreffriction', 
			       'solimpfriction', 
			       'frictionloss', 
			       'margin', 
			       'springlength', 
			       'stiffness', 
			       'damping'}
	_NEW_MUJOCO_ATTR    = {'limited':            bool, 
			       'actuatorfrclimited': bool, 
			       'range':              np.ndarray, 
			       'actuatorfrcrange':   np.ndarray, 
			       'frictionloss':       float, 
			       'width':              float, 
			       'stiffness':          float, 
			       'damping':            float, 
			       'armature':           float}
	_NEW_DEFAULT_VALS   = {'limited':            None, 
			       'actuatorfrclimited': None, 
			       'range':              np.array([0., 0.], dtype=np.float32), 
			       'actuatorfrcrange':   np.array([0., 0.], dtype=np.float32), 
			       'frictionloss':       0., 
			       'width':              0.003, 
			       'stiffness':          0., 
			       'damping':            0., 
			       'armature':           0.}
	_NEW_BLUEPRINT_ATTR = {'limited':            bool,
			       'act_force_limited':  bool, 
			       'min_length':         float,
			       'max_length':         float, 
			       'min_act_force':      float,
			       'max_act_force':      float, 
			       'frictionloss':       float, 
			       'width':              float, 
			       'stiffness':          float, 
			       'damping':            float, 
			       'armature':           float,}



class InfoLaserType(SiteSensorType):

	"""
	This abstract base class is a used to represent the InfoLaser class in type hints. For a detailed
	description see :class:`blueprints.sensors.InfoLaser`.
	"""

	_DIMENSION          = -1
	_NEW_BLUEPRINT_ATTR = {'axis': str}
	_NEW_DEFAULT_VALS   = {'axis': np.array([0., 0., 1.], dtype=np.float32)}


class ActuatorType(CyclicalThingType, NodeThingType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.BaseActuator`.
	"""

	_NEW_DEFAULT_VALS   = {'ctrllimited':  None,
			       'forcelimited': None,
			       'actlimited':   None,
			       'ctrlrange':    np.array([0., 0.], dtype=np.float32),
			       'forcerange':   np.array([0., 0.], dtype=np.float32),
			       'actrange':     np.array([0., 0.], dtype=np.float32),
			       'lengthrange':  np.array([0., 0.], dtype=np.float32),
			       'gear':         np.array([1., 0., 0., 0., 0., 0.], dtype=np.float32),
			       'cranklength':  0.,
			       'dyntype':      'none',
			       'gaintype':     'fixed',
			       'biastype':     'none',
			       'dynprm':       np.array(1., dtype=np.float32),
			       'gainprm':      np.array(1., dtype=np.float32),
			       'biasprm':      np.array(1., dtype=np.float32)}
	_NEW_BLUEPRINT_ATTR = {'ctrllimited':  bool,
			       'forcelimited': bool,
			       'actlimited':   bool,
			       'ctrlrange':    np.ndarray,
			       'forcerange':   np.ndarray,
			       'actrange':     np.ndarray,
			       'lengthrange':  np.ndarray,
			       'gear':         np.ndarray,
			       'cranklength':  float,
			       'dyntype':      str,
		     	       'gaintype':     str,
		     	       'biastype':     str,
		     	       'dynprm':       np.ndarray,
		     	       'gainprm':      np.ndarray,
		     	       'biasprm':      np.ndarray,
		     	       'inparent':     bool}
	_NEW_MUJOCO_ATTR    = {'ctrllimited':  bool,
			       'forcelimited': bool,
			       'actlimited':   bool,
			       'ctrlrange':    np.ndarray,
			       'forcerange':   np.ndarray,
			       'actrange':     np.ndarray,
			       'lengthrange':  np.ndarray,
			       'gear':         np.ndarray,
			       'cranklength':  float,
			       'dyntype':      str,
		     	       'gaintype':     str,
		     	       'biastype':     str,
		     	       'dynprm':       np.ndarray,
		     	       'gainprm':      np.ndarray,
		     	       'biasprm':      np.ndarray}
	_OTHER_REFERENCES   = {'refsite':      'ref_actuators'}
	_PARENT_REFERENCE   =  'refsite'
	_MUJOCO_OBJ         =  'actuator'


class PositionType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.Position`.
	"""

	_NEW_DEFAULT_VALS   = {'kp': 1.}
	_NEW_BLUEPRINT_ATTR = {'kp': float}
	_NEW_MUJOCO_ATTR    = {'kp': float}


class VelocityType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.Velocity`.
	"""

	_NEW_DEFAULT_VALS   = {'kv': 1.}
	_NEW_BLUEPRINT_ATTR = {'kv': float}
	_NEW_MUJOCO_ATTR    = {'kv': float}


class IntVelocityType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.IntVelocity`.
	"""

	_NEW_DEFAULT_VALS   = {'kp': 1.}
	_NEW_BLUEPRINT_ATTR = {'kp': float}
	_NEW_MUJOCO_ATTR    = {'kp': float}


class DamperType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.Damper`.
	"""

	_NEW_DEFAULT_VALS   = {'kv': 1.}
	_NEW_BLUEPRINT_ATTR = {'kv': float}
	_NEW_MUJOCO_ATTR    = {'kv': float}


class CylinderType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.Cylinder`.
	"""

	_NEW_DEFAULT_VALS   = {'timeconst': 1.,
			       'area':      1.,
			       'diameter':  None,
			       'bias': np.zeros(3, dtype=np.float32)}
	_NEW_BLUEPRINT_ATTR = {'timeconst': float,
			       'area':      float,
			       'diameter':  float|None,
			       'bias':      np.ndarray}
	_NEW_MUJOCO_ATTR    = {'timeconst': float,
			       'area':      float,
			       'diameter':  float|None,
			       'bias':      np.ndarray}


class MuscleType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.Muscle`.
	"""

	_NEW_DEFAULT_VALS   = {'timeconst': np.array([0.01, 0.04], dtype=np.float32),
			       'tausmooth': 0.,
			       'range':     np.array([0.75, 1.05], dtype=np.float32),
			       'force':    -1.,
			       'scale':     200,
			       'lmin':      0.5,
			       'lmax':      1.6,
			       'vmax':      1.5,
			       'fpmax':     1.3,
			       'fvmax':     1.2}
	_NEW_BLUEPRINT_ATTR = {'timeconst': np.ndarray,
			       'tausmooth': float,
			       'range':     np.ndarray,
			       'force':     float,
			       'scale':     float,
			       'lmin':      float,
			       'lmax':      float,
			       'vmax':      float,
			       'fpmax':     float,
			       'fvmax':     float}
	_NEW_MUJOCO_ATTR    = {'timeconst': np.ndarray,
			       'tausmooth': float,
			       'range':     np.ndarray,
			       'force':     float,
			       'scale':     float,
			       'lmin':      float,
			       'lmax':      float,
			       'vmax':      float,
			       'fpmax':     float,
			       'fvmax':     float}


class AdhesionType(ActuatorType):

	"""
	This abstract base class is a used to represent the Actuator class in type hints. For a detailed
	description see :class:`blueprints.actuators.Adhesion`.
	"""

	_NEW_DEFAULT_VALS   = {'gain': 1.}
	_NEW_BLUEPRINT_ATTR = {'gain': float}
	_NEW_MUJOCO_ATTR    = {'gain': float}


class LightType(CyclicalThingType, MoveableThingType):

	"""
	This abstract base class is a used to represent the Light class in type hints. For a detailed
	description see :class:`blueprints.light.Light`.
	"""

	_NEW_DEFAULT_VALS   = {'pos':         np.array([ 0., 0., 0.], dtype=np.float32),
			       'dir':         np.array([ 0., 0.,-1.], dtype=np.float32),
			       'attenuation': np.array([ 1., 0., 0.], dtype=np.float32),
			       'ambient':     np.array([ 0., 0., 0.], dtype=np.float32),
			       'diffuse':     np.array([ .7, .7, .7], dtype=np.float32),
			       'specular':    np.array([ .3, .3, .3], dtype=np.float32),
			       'mode':        'fixed',
			       'directional': False,
			       'castshadow':  True,
			       'active':      True,
			       'cutoff':      3.1415926535897932384626433832/2,
			       'exponent':    10.}
	_NEW_BLUEPRINT_ATTR = {'dir':         np.ndarray,
			       'attenuation': np.ndarray,
			       'ambient':     np.ndarray,
			       'diffuse':     np.ndarray,
			       'specular':    np.ndarray,
			       'mode':        str,
			       'directional': bool,
			       'castshadow':  bool,
			       'active':      bool,
			       'cutoff':      float,
			       'exponent':    float}
	_NEW_MUJOCO_ATTR    = {'dir':         np.ndarray,
			       'attenuation': np.ndarray,
			       'ambient':     np.ndarray,
			       'diffuse':     np.ndarray,
			       'specular':    np.ndarray,
			       'mode':        str,
			       'directional': bool,
			       'castshadow':  bool,
			       'active':      bool,
			       'cutoff':      float,
			       'exponent':    float}
	_DEL_MUJOCO_ATTR    = {'euler'}
	_MODES              = {'fixed',
			       'track',
			       'trackcom',
			       'targetbody',
			       'targetbodycom'}
	_OTHER_REFERENCES   = {'target': 'targeting_lights'}
	_PARENT_REFERENCE   =  'target'
	_MUJOCO_OBJ         =  'light'
	_MUJOCO_DATA        =  'light'


class CameraType(CyclicalThingType):

	"""
	This abstract base class is a used to represent the Camera class in type hints. For a detailed
	description see :class:`blueprints.camera.Camera`.
	"""

	_NEW_DEFAULT_VALS   = {'mode':        'fixed',
			       'resolution':  [1, 1], 
			       'fovy':        45.0,
			       'ipd':         0.068}
	_NEW_BLUEPRINT_ATTR = {'mode':        str,
			       'resolution':  np.ndarray, 
			       'fieldofview': float,
			       'ipd':         float, 
			       'fps':         int}
	_NEW_MUJOCO_ATTR    = {'mode':        str,
			       'resolution':  np.ndarray, 
			       'fovy':        float,
			       'ipd':         float}
	_MODES              = {'fixed',
			       'track',
			       'trackcom',
			       'targetbody',
			       'targetbodycom'}
	_OTHER_REFERENCES   = {'target': 'targeting_cameras'}
	_PARENT_REFERENCE   =  'target'
	_MUJOCO_OBJ         =  'camera'


class AgentType(BodyType):

	"""
	This abstract base class is a used to represent the Agent class in type hints. For a detailed
	description see :class:`blueprints.agent.Agent`.
	"""



"""
The Typechecker class is used to restrict the argument types of functions and methods. 
If a function or method is decoraded with TypeChecker.restrict its arguments and returned 
values are checked against the type hints, if provided and a TypeError is raised if they 
are violated.
"""
import sys
import inspect
import traceback
from functools import wraps
from types import GenericAlias, UnionType, TracebackType, FunctionType, MethodType



class ArgumentError(Exception):

	"""
	This class is used to construct the error message for restricted type violations and 
	to be caught in the methods ``@restrict`` wrapper call as distinct from other errors that 
	might have occured further down the traceback.
	"""
	
	def __init__(self, arg, arg_type, arg_name, name):
		"""
		The ArgumentError is used to construct the Error message for a restricted type violation.
		
		Parameters
		----------
		arg : object
			The argument given to the function.
		arg_type : type
			The type as restricted by the function signature.
		arg_name : str
			The name of the argument from the function signature.
		name : str
			The name of the method or function called.
		"""
		arg_str = str(arg)
		if arg_str.count('\n') > 8:
			arg_lines = arg_str.split()
			arg_str = '\n'.join(arg_lines[:4]) + '\n[…]\n' + '\n'.join(arg_lines[-4:])
		elif len(arg_str) > 1024:
			arg_str = arg_str[:100] + '\n[…]\n' + arg_str[-100:]
		super().__init__(f"""
A type restriction in function {name} was violated: 
{arg_name} is supposed to be of type {arg_type}, but got {type(arg)} instead.
The recieved argument for {arg_name} was:
{arg_str}
""")



class TypeChecker:

	"""
	TypeChecker controls function and class methods arguments for their validity.
	If the given type of an argument violates the type hint, a TypeError is raised.
	Use TypeChecker.restrict as a decorator to restrict a function or methods types. 
	"""
	
	@classmethod
	def restrict(cls, 
		     func: FunctionType|MethodType):
		"""
		Used as a decorator it ansurses that the decorated methods or functions
		arguments are type restricted according to the arguments type hints if provided.
		You can use None as an optional type hint instead of Nonetype.
		
		Parameters
		----------
		func : FunctionType | MethodType
			A function or method to be type restricted.
		
		Returns
		-------
		TYPE
			A wrapped function that resitrcts argument and return value types and raises a TypeError if violated.
		"""
		if not hasattr(func, '__code__') or not hasattr(func, '__annotations__'):
			return func
		code        = func.__code__
		annotations = func.__annotations__
		var_names   = code.co_varnames
		n_args      = code.co_argcount
		arg_names   = var_names[:n_args]
		arg_types   = [annotations[arg] for arg in arg_names if arg in annotations]
		static      = isinstance(func, (staticmethod, property, classmethod))
		if 'return' in annotations:
			return_type = annotations['return']
		else:
			return_type = False
		@wraps(func)
		def wrapper(*args, **kwargs):
			try:
				for arg, arg_type, arg_name in zip(args[0 if static else 1:], arg_types, arg_names[0 if static else 1:]):
					cls.__validate(arg=arg, 
						       arg_name=arg_name, 
						       arg_type=arg_type, 
						       name=func.__qualname__)
				for arg_name, arg in kwargs.items():
					if arg_name not in arg_names or arg_name not in annotations:
						continue
					cls.__validate(arg=arg, 
						       arg_name=arg_name, 
						       arg_type=annotations[arg_name], 
						       name=func.__qualname__)
				result = func(*args, **kwargs)
				if not return_type is False: # DO NOT REDUCE TO if not return_type!
					cls.__validate(arg=result, 
						       arg_name='the returned value', 
						       arg_type=return_type, 
						       name=func.__qualname__)
				return result
			except ArgumentError as error:
				raise TypeError(error) from None
			except Exception as error:
				tb  = error.__traceback__
				skip = lambda tb: tb.tb_frame.f_globals['__file__'] == __file__
				tbs = []
				while tb is not None:
					if not skip(tb):
						tbs.append(tb)
					tb = tb.tb_next
				if tbs and not skip(tbs[-1]):
					tbs[-1].tb_next = None
				for tb_A, tb_B in zip(tbs[:-1], tbs[1:]):
					tb_A.tb_next = tb_B
				if tbs:
					error.__traceback__ = tbs[0]
				raise error

		wrapper.__annotations__ = annotations
		if hasattr(func, __doc__):
			wrapper.__doc__ = func.__doc__
		return wrapper


	@classmethod
	def __validate(cls, 
		       arg:      object, 
		       arg_name: str, 
		       arg_type: type, 
		       name:     str):
		"""
		Helper method for type validation.
		
		Parameters
		----------
		arg : object
			An argument to bec checked.
		arg_name : str
			The name of he argument to be checked.
		arg_type : type
			The type convention the argument must abide by.
		name : str
			The name of the function or method that is validated.
		
		Raises
		------
		ArgumentError
		Description
		
		No Longer Raises
		----------------
		TypeError
		If the arg_types are violated a TypeError is raised.
		"""
		if not cls.__validate_type(arg, arg_type):
			raise ArgumentError(arg, arg_type, arg_name, name)


	@classmethod
	def __validate_type(cls, 
			    arg:      object, 
			    arg_type: type) -> bool:
		"""
		Helper method to check whether the arg_type is vioated.
		
		Parameters
		----------
		arg : object
			The argument to be checked.
		arg_type : type
			One or multiple argument types to be checked against.
		
		Returns
		-------
		bool
			Indicates wether the arg_type was valid according to arg_type.
		"""
		if isinstance(arg_type, GenericAlias):
			if not hasattr(arg, '__iter__'):
				return False
			seq_valid   = isinstance(arg, arg_type.__origin__)
			unroll = lambda x, types: any(cls.__validate_type(x, single) for single in types)
			items_valid = all(map(lambda x, types=arg_type.__args__: unroll(x, types), arg))
			return seq_valid and items_valid
		elif isinstance(arg_type, UnionType):
			return any(cls.__validate_type(arg, sub_arg_type) for sub_arg_type in arg_type.__args__)
		elif arg_type is None:
			return arg is None
		else:
			return isinstance(arg, arg_type)



restrict = TypeChecker.restrict



if __name__ == '__main__':
	@TypeChecker.restrict
	def f(x: (int, float), 
	      k: set[int] | dict, 
	      l: list[int | float], 
	      s: str='string', 
	      **kwargs) -> None | int:
		"""Summary
		
		Parameters
		----------
		x : int, float
			Description
		k : set[int] | dict
			Description
		l : list[int | float]
			Description
		s : str, optional
			Description
		**kwargs
			Description
		
		Returns
		-------
		None | int
		"""
		m = None
		print('hi')
		return 0
	def g():
		"""Summary
		"""
		f(3.2, k=0, l=[1, 2.3, 0])
	g()
	


"""
.. note::
	If you found yourself here not being a dev, you should probably :attr:`skip this section <BaseThing.ID>` or 
	look up the correspinging docs page of the Thing you were looking after for which this module only 
	implements its corresponding mirror types. A link to the docs page is to be found in the Types doc 
	string.
"""


import blueprints as blue
from blueprints import restrict

import numpy as np
import inspect
import xml.etree.ElementTree as xml
from copy import copy
from collections import defaultdict



class BaseThing(blue.ThingType):

	"""
	BaseThing implements the basic attributes that are used by all Things. 
	
	All Things have certain attributes that are included in the mujoco XML construction. 
	The names of those attributes can be obtained from the :attr:`_MUJOCO_ATTR` property 
	which is dictionary with attribute names as keys and the types their value will have 
	in the Thing instance. If a new Thing class is defined that inherits from BaseThing that 
	introduces new attributes to be written to xml, they can be added by defining a class 
	property ``_NEW_MUJOCO_ATTR`` in the same structure that ``_MUJOCO_ATTR`` is defined.
	``_MUJOCO_ATTR`` then simply aggregates the new additions from ``_NEW_MUJOCO_ATTR``. 
	If a new Thing class inherits a mujoco attribute, that is no longer used, it can be 
	deleted from ``_MUJOCO_ATTR`` by defining it in a class property ``_DEL_MUJOCO_ATTR``, 
	must be a set of the names to be detached.
	
	When a Thing is copied, all attributes needed to reconstruct it are retrieved to create
	a new instance. For this the :attr:`blueprints.type.ThingType._BLUEPRINT_ATTR` property is used. New attributes 
	that are feed into the inheriting ``Thing.__init__`` are registered in the class property 
	``_NEW_BLUEPRINTS_ATTR`` and old attributes that are no longer used are registered in 
	``_DEL_BLUEPRINTS_ATTR``. Their structures are analogous to the ``_MUJOCO_ATTR``.
	
	If a Thing gets copied all attributes necessary to instantiate the copy are themselves 
	copied to ensure that the Thing and its copy do not share a reference to the same object
	in any attribute. However, if an attribute is added to :attr:`blueprints.type.ThingType._NO_COPY_VALS` it is not 
	copied. New attributes to be excluded from being copied can be added by setting the class
	property ``_NEW_NO_COPY_VALS`` to the classes Type as a set of attribute names.
	
	If a Thing is reconstructed from an xml string, some attribute in the xml tag are not 
	attributes of the Thing class. For example, a geom xml tag will have a type defining
	its shape ``<geom type='sphere'>`` but the corresponding :class:`blueprints.geoms.Sphere`
	class specifies the type only implicitly in its name. To retrieve the correct class in
	reconstruction from xml then, those implicit attributes are gathered in the class 
	property :attr:`blueprints.type.ThingType._DERIVED_ATTR` to explicitly derived later. Adding new attributes to it
	is done by setting the class property ``_NEW_DERIVED_ATTR`` of the Things Type as a set of attribute 
	names.

	To ensure readability of the resulting xml, attributes that are set to their mujoco 
	defaults are omitted from the xml representations (even though all attributes are set
	in the Thing instance). To check whether an attributes value equals its default, it can
	be obtained from the ``_DEFAULT_VALS`` class property, which is structured as a 
	dictionary of attribute names as keys and their defaults as values. New default values 
	of inheriting classes are added analogously to the previous cases by setting them in a 
	class property ``_NEW_DEFAULT_VALS`` of its Type.
	
	Parameters
	----------
	name : str | None, optional
		This is the name for the object specified by the user. It differs from other 
		user specified properties, in that it might be altered without the users knowledge
		to resolve name conflicts.
	parent : blue.NodeThingType | None, optional
		The parent to which the Thing is attach, if it unattached parent is None.
	**kwargs
		The aggregation of keyword arguments that have not yet been caught by other inheriting 
		Things `__init__` are not used and serve as a dummy variable.
	
	Attributes
	----------
	ID : int
		Each Thing has a unique ID that is gets on initialization from the global REGISTER.
	"""
	
	@restrict
	def __init__(self, 
		     name:   str|None = None, 
		     parent: blue.NodeThingType|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		name : str | None, optional
			This is the name for the object specified by the user. It differs from other 
			user specified properties, in that it might be altered without the users knowledge
			to resolve name conflicts.
		parent : blue.NodeThingType | None, optional
			The parent to which the Thing is attach, if it unattached parent is None.
		**kwargs
			The aggregation of keyword arguments that have not yet been caught by other inheriting 
			Things `__init__` are not used and serve as a dummy variable.
		"""
		self.ID           = blue.REGISTER.get_ID()
		self._name_scope  = None
		self._name        = name or f'anonymous_{self.__class__.__name__.lower()}'
		self.parent       = parent


	#def __setattr__(self, attr, value):
		"""
		If the Thing is attach to a :class:`blueprints.World` it is informed that its child 
		was altered such that on the next method call on the World, that require the current 
		build, that it needs to rebuild.
		
		Parameters
		----------
		attr : str
			The name of the attribute to be set.
		value : object
			The value assigned to the attribute.
		"""
		#root = self.root
		#if isinstance(root, blue.WorldType) and not isinstance(self, blue.WorldType):
		#	root.built = False
		#super().__setattr__(attr, value)


	def __str__(self) -> str:
		"""
		Returns
		-------
		str
			The representation of a Thing is the xml tag. Children are not included and not 
			all properties are necessarily represented in the xml tag, either if they are 
			located in an external Thing as in :class:`blueprints.geoms.Mesh` or is they are 
			set to default values in which case they are omitted from xml to ensure readability.
		"""
		if hasattr(self, '_MUJOCO_OBJ'):
			xml_element = xml.Element(self._MUJOCO_OBJ, **self._mujoco_specs())
			return xml.tostring(xml_element, encoding='unicode')
		else:
			return repr(self)


	def __repr__(self) -> str:
		"""
		Returns
		-------
		str
			A Thing is represented starting with its class name followed by its name property and 
			its ID.
		"""
		return f'{self.__class__.__name__}<{self.name}:{self.ID}>'


	@restrict
	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		self._xml_root = xml.SubElement(parent, 
						self._MUJOCO_OBJ, 
						**self._mujoco_specs(kwargs))
		return self._xml_root


	@restrict
	def copy(self, **kwargs) -> blue.ThingType:
		"""
		This method constructs a copy of the Thing with possible alterations to its 
		attributes as specified in kwargs.
		
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to the Thing.__init__ to replace the attributes 
			of the current Thing from which copy is called.
		
		Returns
		-------
		blue.ThingType
			A new instance of the Thing
		"""
		blueprint_specs = self._blueprint_specs()
		blueprint_specs.update(kwargs)
		#print('kw', kwargs, blueprint_specs['pos'])
		if 'name' not in kwargs:
			blueprint_specs['name'] = self._name
		thing = self.__class__(**blueprint_specs)
		return thing


	def _clear_step_cache(self):
		"""
		Clears the cached properties derived from simulation values after each time step.
		"""
		for attr in self._STEP_CACHE():
			self.__delattr__(attr)


	@property
	def _location_range(self) -> tuple[np.ndarray, np.ndarray]|None:
		"""
		This is a dummy function that is used by other Things that have a physical size and 
		location to determine the range of space it occupies.
		
		Returns
		-------
		None | None
			None
		"""
		return None


	@property
	def _launched(self):
		"""
		This property indicates whether the world the Thing belongs to has been build 
		and a simulation is launched.

		Returns
		-------
		bool
		"""
		root = self.root
		return isinstance(root, blue.WorldType) and root._built

	# XML INITIALIZATION METHODS

	@restrict
	@classmethod
	def _xml_element_args(cls, xml_element: xml.Element) -> tuple:
		"""
		This method extracts the attributes from xml elements and converts them into 
		a type compatible to the Things corresponding attribute.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which attributes are reconstructed.
		
		Returns
		-------
		tuple
			Three dictionaries are returned, the `init_args` containing arguments passed 
			to :meth:`Thing.__init__`, the `post_args` containing the arguments that are 
			set via :meth:`Thing.__setattr__` after initialization and `rest_args` 
			containing those arguments which must be handled manually.
		"""
		init_args = dict()
		post_args = dict()
		rest_args = dict()
		for key, val in xml_element.items():
			if key in cls._DERIVED_ATTR():
				arg_type = cls._DERIVED_ATTR()[key]
				rest_args[key] = cls._convert_from_string(val, arg_type)
			elif key in cls._BLUEPRINT_ATTR():
				arg_type = cls._BLUEPRINT_ATTR()[key]
				init_args[key] = cls._convert_from_string(val, arg_type)
			elif key in cls._MUJOCO_ATTR():
				arg_type = cls._MUJOCO_ATTR()[key]
				if not val:
					continue
				post_args[key] = cls._convert_from_string(val, arg_type)
			else:
				rest_args[key] = val
		return init_args, post_args, rest_args


	@restrict
	@classmethod
	def _from_xml_element(cls, xml_element: xml.Element) -> blue.ThingType:
		"""
		This method reconstructs a Thing from an xml element. If any argument for 
		an inheriting class has to be set manually this method must be overwritten.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a Thing is reconstructed.
		
		Returns
		-------
		blue.ThingType
			The reconstructed Thing.
		"""
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['copy'] = False
		obj = object.__new__(cls)
		obj.__init__(**init_args)
		for key, val in post_args.items():
			setattr(obj, key, val)
		return obj


	@restrict
	def _mujoco_specs(self, specs: dict|None = None) -> dict:
		"""
		This method assists the construction of the Things xml representations. 
		It constructs a dictionary containing the names of the attributes as keys 
		and the string representations of their values.
		
		Returns
		-------
		dict
			A dictionary from which the xml tags attributes are build.
		
		Parameters
		----------
		specs : dict | None, optional
			Specs can be set as a dictionary to overwrite some attributes with custom 
			values.
		"""
		specs        = specs or {}
		condition    = lambda name, attr: attr is not None and (name not in self._DEFAULT_VALS() or np.any(attr != self._DEFAULT_VALS()[name]))
		#convert      = lambda x: str(x) if not isinstance(x, np.ndarray) else self._numpy_to_string(x)
		#convert      = lambda x: str(x).lower() if not isinstance(x, np.ndarray) else self._numpy_to_string(x)
		convert      = self._convert_to_string
		mujoco_attrs = map(lambda name: (name, self.__getattribute__(name)), self._MUJOCO_ATTR().keys())
		mujoco_specs = {name: convert(attr) for name, attr in mujoco_attrs if condition(name, attr)}
		conv_specs   = {name: convert(attr) for name, attr in specs.items()}
		mujoco_specs.update(conv_specs)
		for spec, val in specs.items():
			if val is None:
				del mujoco_specs[spec]
		return mujoco_specs


	@restrict
	def _blueprint_specs(self, specs: dict|None = None) -> dict:
		"""
		This method is used to assist the copying of a Thing. It obtains all attributes 
		from the Thing and potentially copies them to avoid multiple Things referencing 
		the same attribute value.
		
		Returns
		-------
		dict
			The returned dictionary has the attribute names as keys and their corresponding
			values.
		
		Parameters
		----------
		specs : dict | None, optional
			Specs can be specified as a dictionary to overwrite some attributes with custom 
			values.
		"""
		specs           = specs or {}
		#condition       = lambda attr, name: True
		condition       = lambda name, attr: attr is not None and (name not in self._DEFAULT_VALS() or np.any(attr != self._DEFAULT_VALS()[name]))
		convert         = lambda name, attr: attr if name in self._NO_COPY_ATTR() else copy(attr)
		blueprint_attrs = map(lambda name: (name, self.__getattribute__(name)), self._BLUEPRINT_ATTR().keys())
		#converted_attrs = {name: convert(name, attr) for name, attr in blueprint_attrs if condition(name, attr)}
		blueprint_specs = {name: convert(name, attr) for name, attr in blueprint_attrs if condition(name, attr)}
		#blueprint_specs = {name: attr if name not in self._COPY_BLUE_ATTR else attr.copy() for name, attr in converted_attrs.items()}
		blueprint_specs.update(specs)
		for spec, val in specs.items():
			if val is None:
				del blueprint_specs[spec]
		return blueprint_specs

	# XML CONVERSION METHODS

	@restrict
	@classmethod
	def _convert_to_string(cls, obj) -> str:
		"""
		This method converts the string representation of a value from an xml 
		attribute into its corresponding blueprints attribute.
		
		Parameters
		----------
		string : str
			value representation 
		dtype : TYPE
			data type to which the value is to be converted
		
		Returns
		-------
		object
			The returned value can be assigned to the corresponding Things attribute.
		"""
		if isinstance(obj, bool):
			return cls._bool_to_string(obj)
		elif isinstance(obj, np.ndarray):
			return cls._numpy_to_string(obj)
		else:
			return str(obj)

	@restrict
	@classmethod
	def _convert_from_string(cls, string: str, dtype) -> object:
		"""
		This method converts the string representation of a value from an xml 
		attribute into its corresponding blueprints attribute.
		
		Parameters
		----------
		string : str
			value representation 
		dtype : TYPE
			data type to which the value is to be converted
		
		Returns
		-------
		object
			The returned value can be assigned to the corresponding Things attribute.
		"""
		if dtype == bool:
			return cls._string_to_bool(string)
		elif dtype == np.ndarray:
			return cls._string_to_numpy(string)
		else:
			return dtype(string)


	@restrict
	@staticmethod
	def _string_to_numpy(string: str) -> np.ndarray:
		"""
		Convertes a string representation of a value to np.ndarray.
		
		Parameters
		----------
		string : str
			value representation
		
		Returns
		-------
		np.ndarray
			reconstructed np.ndarray
		"""
		sep = ',' if ',' in string else ' '
		return np.array(list(map(float, string.split(sep))))


	@restrict
	@staticmethod
	def _numpy_to_string(array: np.ndarray) -> str:
		"""
		Convertes a np.ndarray to a string representation of its value.
		
		Parameters
		----------
		array : np.ndarray
			any np.ndarray
		
		Returns
		-------
		str
			reconstructed string representation
		"""
		return str(array) if array.shape == () else ' '.join(map(str, array))


	@restrict
	@staticmethod
	def _string_to_bool(string: str) -> bool|None:
		"""
		Convertes a string representation of a value boolean.
		
		Parameters
		----------
		string : str
			value representation
		
		Returns
		-------
		bool
			reconstructed boolean
		
		Raises
		------
		ValueError
			If the string is not 'false', 'true' in any capitalization an Error is raised
		"""
		clean_string = string.lower().strip()
		if clean_string == 'true':
			return True
		elif clean_string == 'false':
			return False
		elif clean_string == 'auto':
			return None
		else:
			raise ValueError(f"_string_to_bool takes either 'true' or 'false' in any capitalization as argument, got {string}")


	@restrict
	@staticmethod
	def _bool_to_string(value: bool|None) -> str:
		"""
		Converts a boolean to a string representation of its value.
		
		Parameters
		----------
		value : bool
			any boolean value
		
		Returns
		-------
		str
			reconstructed string representation
		"""
		if value is None:
			return 'auto'
		return str(value).lower()

	# KINEMATIC TREE PROPERTIES

	@property
	def root(self) -> blue.ThingType:
		"""
		The root attribute is immutable and changes only if the structure of the kinematic tree is 
		altered.

		Returns
		-------
		blue.ThingType
			Retrieves the root of the kinematic tree the Thing in part of, possibly 
			the Thing itself if it has no parent.
		"""
		root = self
		parent = getattr(root, '_parent', None)
		while parent is not None:
			root = parent
			parent = getattr(root, '_parent', None)
		return root


	@property
	def path(self) -> list:
		"""
		The path attribute is immutable and changes only if the structure of the kinematic tree is 
		altered.

		Returns
		-------
		list
			Retrieves a list of all Things that lead from the Thing itself to its root.
		"""
		node = self
		path = [node]
		parent = getattr(node, '_parent', None)
		while parent is not None:
			node = parent
			path.append(node)
			parent = getattr(node, '_parent', None)
		path.reverse()
		return path


	@property
	def parent(self) -> blue.ThingType | None:
		"""
		The parent attribute is mutable. However it is recommended to use the :meth:`NodeThing.attach` method to alter 
		the kinematic tree.

		Returns
		-------
		blue.ThingType | None
			If the Thing is attach in a kinematic tree, its parent is returned else None.
		"""
		if hasattr(self, '_parent'):
			return self._parent
		else:
			return None


	@parent.setter
	@restrict
	def parent(self, parent: blue.ThingType|None) -> None:
		"""
		The parent attribute is mutable. However it is recommended to use the :meth:`NodeThing.attach` method to alter 
		the kinematic tree.

		Parameters
		----------
		parent : blue.ThingType | None
			The parent references another Thing in a kinematic tree to which the Thing 
			is attached. Setting this property manually is generally not recommended since 
			failure to also attach it to its new parent may lead to unintended consequences.
			Modifications of the kinematic tree should be done via the :meth:`NodeThing.attach` 
			or the :meth:`NodeThing.detach` methods instead.
		"""
		self._parent = parent

	# MUJOCO PROPERTIES

	@property
	def name(self) -> str:
		"""
		The returned name might differ from the user specification by an enumerations 
		scheme that is applied if two Things of the same type in a kinematic tree share 
		the same name.
		
		Returns
		-------
		str
			possibly extended name of the Thing
		"""
		if self._name_scope is not None:
			return self._name_scope.name(self)
		else:
			return self._name


	@name.setter
	@restrict
	def name(self, name: str):
		"""
		Parameters
		----------
		name : str
			The name assigned to the Thing might be altered by enumeration in the case of 
			a naming conflict.
		"""
		if self._name_scope is not None:
			#self._name_scope.assign(self, name)
			assert not self.root._built
			raise Exception("""Name are only allowed to be changed for Things in unbuild Worlds.
To change the of this Thing use :meth:`World.unbuild`.""")
		else:
			self._name = name

	# INHERITANCE PROPERTIES

	@classmethod
	#@property
	def _NO_COPY_ATTR(cls) -> dict:
		"""
		Returns
		-------
		dict
			Attributes in this dictionary will not be copied when the Thing is copied.
		"""
		if hasattr(cls, '_NEW_NO_COPY_ATTR'):
			NO_COPY_ATTR = cls._NEW_NO_COPY_ATTR.copy()
		else:
			NO_COPY_ATTR = set()
		for base in cls.__bases__:
			if hasattr(base, '_NO_COPY_ATTR'):
				NO_COPY_ATTR.update(base._NO_COPY_ATTR())
		return NO_COPY_ATTR


	@classmethod
	#@property
	def _SINGLE_CHILD_ATTR(cls) -> dict:
		"""
		Returns
		-------
		dict
			Attributes in this dictionary will not be copied when the Thing is copied.
		"""
		if hasattr(cls, '_NEW_SINGLE_CHILD_ATTR'):
			SINGLE_CHILD_ATTR = cls._NEW_SINGLE_CHILD_ATTR.copy()
		else:
			SINGLE_CHILD_ATTR = dict()
		for base in cls.__bases__:
			if hasattr(base, '_SINGLE_CHILD_ATTR'):
				SINGLE_CHILD_ATTR.update(base._SINGLE_CHILD_ATTR())
		return SINGLE_CHILD_ATTR


	@classmethod
	#@property
	def _DERIVED_ATTR(cls) -> dict:
		"""
		Returns
		-------
		dict
			Attributes in this dictionary will have to be derived by the global `REGISTER`.
		"""
		if hasattr(cls, '_NEW_DERIVED_ATTR'):
			DERIVED_ATTR = cls._NEW_DERIVED_ATTR.copy()
		else:
			DERIVED_ATTR = dict()
		for base in cls.__bases__:
			if hasattr(base, '_DERIVED_ATTR'):
				DERIVED_ATTR.update(base._DERIVED_ATTR())
		return DERIVED_ATTR


	@classmethod
	#@property
	def _STEP_CACHE(cls) -> dict:
		"""
		Returns
		-------
		set
			Attributes in this set are cached for each time step and deleted afterwards.
		"""
		if hasattr(cls, '_NEW_STEP_CACHE'):
			STEP_CACHE = cls._NEW_STEP_CACHE.copy()
		else:
			STEP_CACHE = set()
		for base in cls.__bases__:
			if hasattr(base, '_STEP_CACHE'):
				STEP_CACHE.update(base._STEP_CACHE())
		return STEP_CACHE


	@classmethod
	#@property
	def _MUJOCO_ATTR(cls) -> dict:
		"""
		Returns
		-------
		dict
			Attributes from this dictionary are used to construct the xml representation.
		"""
		if hasattr(cls, '_NEW_MUJOCO_ATTR'):
			MUJOCO_ATTR = cls._NEW_MUJOCO_ATTR.copy()
		else:
			MUJOCO_ATTR = dict()
		for base in cls.__bases__:
			if hasattr(base, '_MUJOCO_ATTR'):
				MUJOCO_ATTR.update(base._MUJOCO_ATTR())
		if hasattr(cls, '_DEL_MUJOCO_ATTR'):
			for attr in cls._DEL_MUJOCO_ATTR:
				if attr in MUJOCO_ATTR:
					del MUJOCO_ATTR[attr]
		return MUJOCO_ATTR


	@classmethod
	#@property
	def _BLUEPRINT_ATTR(cls) -> dict:
		"""
		Returns
		-------
		dict
			Attributes from this dictionary are used to copy the Thing.
		"""
		if hasattr(cls, '_NEW_BLUEPRINT_ATTR'):
			BLUEPRINT_ATTR = cls._NEW_BLUEPRINT_ATTR.copy()
		else:
			BLUEPRINT_ATTR = dict()
		for base in cls.__bases__:
			if hasattr(base, '_BLUEPRINT_ATTR'):
				BLUEPRINT_ATTR.update(base._BLUEPRINT_ATTR())
		if hasattr(cls, '_DEL_BLUEPRINT_ATTR'):
			for attr in cls._DEL_BLUEPRINT_ATTR:
				if attr in BLUEPRINT_ATTR:
					del BLUEPRINT_ATTR[attr]
		return BLUEPRINT_ATTR


	@classmethod
	#@property
	def _DEFAULT_VALS(cls) -> dict:
		"""
		Returns
		-------
		dict
			This dictionary stores the default values of all attributes.
		"""
		if hasattr(cls, '_NEW_DEFAULT_VALS'):
			DEFAULT_VALS = cls._NEW_DEFAULT_VALS.copy()
		else:
			DEFAULT_VALS = dict()
		for base in cls.__bases__:
			if hasattr(base, '_DEFAULT_VALS'):
				DEFAULT_VALS.update(base._DEFAULT_VALS())
		return DEFAULT_VALS



"""
Geometry contains primitive functions for geometric operations on MoveableThings 
position and orientation. Orientations are currently only supported for (improper)
Euler angles and angles are currently assumed to be in radian. To convert angles the 
predefined constants ``DEGREES_TO_RADIANS`` and ``RADIANS_TO_DEGREES`` can be used.

Attributes
----------
TAU : float
	Tau is the circumference of a unit circle. :math:`\\pi \\times 2 = \\tau`.
PI : float
	Pi is the half circumference of a unit circle. :math:`\\pi \\times 2 = \\tau`.
DEGREES_TO_RADIANS : float
	To convert an angle in degrees to radian it can be multiplied with this constant.
RADIANS_TO_DEGREES : float
	To convert an angle in radian to degrees it can be multiplied with this constant.
"""
import numpy as np
import blueprints as blue



TAU = 6.2831853071795864769252867664
PI  = 3.1415926535897932384626433832

DEGREES_TO_RADIANS = TAU / 360
RADIANS_TO_DEGREES = 360 / TAU



class Rotation:
	"""
	This class implements multiple functions that operate on a MoveableThings orientation.
	
	Attributes
	----------
	X : np.ndarray
		The unit vector :math:`\\vec{\\text{id}_x}` representing the ``X``-axis.
	Y : np.ndarray
		The unit vector :math:`\\vec{\\text{id}_y}` representing the ``Y``-axis.
	Z : np.ndarray
		The unit vector :math:`\\vec{\\text{id}_z}` representing the ``Z``-axis.
	"""
	X = np.array([1, 0, 0])
	Y = np.array([0, 1, 0])
	Z = np.array([0, 0, 1])
	@classmethod
	def euler(cls, 
		  vector, 
		  alpha, 
		  beta, 
		  gamma):
		"""
		Rotates a vector according to improper euler angles.
		
		Parameters
		----------
		vector : np.ndarray
			The vector to be rotated.
		alpha : float
			The angle along which the vector is rotated on the ``X``-axis.
		beta : float
			The angle along which the vector is rotated on the ``Y``-axis.
		gamma : float
			The angle along which the vector is rotated on the ``Z``-axis.
		
		Returns
		-------
		np.ndarray
			The rotated vector.
		"""
		R = cls.E_rot(alpha, beta, gamma)
		return R @ vector


	@classmethod
	def rotate_around_axis(cls, 
			       vector, 
			       axis, 
			       angle):
		"""
		Rotates a vector around an axis for a given angles.
		
		Parameters
		----------
		vector : np.ndarray
			The vector to be rotated
		axis : np.ndarray
			The axis around which the vector is rotated according to the right hand rule.
		angle : float
			The angle by which the vector is rotated.
		
		Returns
		-------
		np.ndarray
			The rotated vector.
		"""
		axis = Vector.normalize(axis)
		return vector * np.cos(angle) + \
		np.cross(axis, vector) * np.sin(angle) + \
		axis * np.dot(axis, vector) * (1 - np.cos(angle))


	@classmethod
	def global_orientation(cls, node):
		"""
		This method calculates the global orientation of a Thing (typically a node in a kinematic tree).
		
		Parameters
		----------
		node : Thing
			A Thing (typically a node in a kinematic tree) for which the global orientation is calculated w.r.t it's root.
		
		Returns
		-------
		(float, float, float)
			The improper euler angles representing the global orientation.
		"""
		R = np.eye(3)
		while node is not None and isinstance(node, blue.MoveableThingType):
			M = cls.E_rot(node.alpha, node.beta, node.gamma)
			R = M @ R
			node = node.parent
		alpha, beta, gamma = cls.reference_frame_to_euler(R)
		return alpha, beta, gamma


	@classmethod
	def global_rotation_matrix(cls, node):
		"""
		This method calculates the rotation matrix for the global orientation of a Thing (typically a node in e kinematic tree) w.r.t. to its root.
		
		Parameters
		----------
		node : Thing
			A Thing (typically a node in a kinematic tree) for which the global orientation is calculated w.r.t it's root.
		
		Returns
		-------
		np.ndarray
			The global orientation matrix for the node.
		"""
		alpha, beta, gamma = cls.global_orientation(node)
		return cls.E_rot(alpha, beta, gamma)


	@classmethod
	def euler_to_reference_frame(cls, 
				     alpha: float, 
				     beta:  float, 
				     gamma: float):
		"""
		This method constructs a reference frame (i. e. the unit vectors for the axis of the local coordinate system) for (improper) euler angles.
		
		Parameters
		----------
		alpha : float
			The angle by which the ``X``-axis is rotated to reach the reference frame.
		beta : float
			The angle by which the ``Y``-axis is rotated to reach the reference frame.
		gamma : float
			The angle by which the ``Z``-axis is rotated to reach the reference frame.
		
		Returns
		-------
		(np.ndarray, np.ndarray, np.ndarray)
			A tuple of the three unit vectors for the given orientation representing the reference frame.
		"""
		M = cls.M_rot(alpha, beta, gamma)
		R_x, R_y, R_z = np.rollaxis(M, axis=1)
		return R_x, R_y, R_z


	@classmethod
	def reference_frame_to_euler(cls, R):
		"""
		Computes the (improper) euler angles that lead to the given reference frame.
		
		Parameters
		----------
		R : np.ndarray
			A matrix representing the reference frame as a concatenation of the unit coulmn vectors of the orientations coordinate system.
		
		Returns
		-------
		(float, float, float)
			The (improper) euler angles for the given reference frame.
		
		Raises
		------
		ArithmeticError
		If the reference frame does not contain orthogonal vectors that do not adhere to the right hand rule convention am ArithmeticError is raised.
		"""
		X = np.array([1, 0, 0])
		Y = np.array([0, 1, 0])
		Z = np.array([0, 0, 1])
		R_x, R_y, R_z = np.rollaxis(R, axis=1)
		try:
			assert Vector.equal(np.cross(R_x, R_y), R_z)
		except AssertionError:
			raise ArithmeticError('Referenceframe to euler convertion only supports rotation and not reflections but the given reference frame contains a reflected axis.')
		R_x = Vector.normalize(R_x)
		R_y = Vector.normalize(R_y)
		R_z = Vector.normalize(R_z)
		R_x1, R_x2, R_x3 = R_x
		R_y1, R_y2, R_y3 = R_y
		R_z1, R_z2, R_z3 = R_z
		if R_y1 != 0:
			Y_bar = np.array([0, -R_z3/R_y1, R_z2/R_y1])
			Y_bar = Vector.normalize(Y_bar)
			X_bar = np.cross(Y_bar, R_z)
		else:
			X_bar = R_x
			Y_bar = R_y
		Z_bar = np.cross(X, Y_bar)
		alpha = cls.angle(Z,     Z_bar, axis=X)
		beta  = cls.angle(X,     X_bar, axis=Y_bar)
		gamma = cls.angle(Y_bar, R_y,   axis=R_z)
		return alpha, beta, gamma


	@classmethod
	def angle(cls, a, b, axis=None) -> float:
		"""
		Computes the angle between two vectors. 
		If axis is set, the angle is measured according to a rotation 
		around axis from a to b, otherwise the shortest angle is returned.
		
		Parameters
		----------
		a : np.ndarray
			A vector for which the angle is computed w.r.t. b.
		b : np.ndarray
			A vector for which the angle is computed w.r.t. a.
		axis : None|np.ndarray, optional
			The axis of rotation can be specified optionally, if this is not done, the resulting angles will be in the range [0, PI).
		
		Returns
		-------
		float
			The resulting angle will be in the range [0, PI) if axis is not specified, otherwise the angle will be in the range [0, TAU).
		"""
		a     = Vector.normalize(a)
		b     = Vector.normalize(b)
		c     = np.cross(a, b)
		axis  = Vector.normalize(axis) if axis is not None else c
		sign  = np.sign(axis @ c)
		sign  = sign if sign != 0 else 1
		# DOT PRODUCT CLIPPING FOR NUMERICAL STABILITY
		angle = np.arccos(np.clip(a @ b, -1, 1))
		return sign * angle


	@classmethod
	def N_rot(cls, node=None) -> np.ndarray:
		"""
		This method returns the rotation matrix for a given Thing. If the Thing does not inherit
		from MoveableThing the identity metrix is returned.
		
		Parameters
		----------
		node : None, ThingType
			A Thing, typically a node from a kinematic tree.
		
		Returns
		-------
		np.ndarray
			A rotation matrix from the Things orientation.
		"""
		if node and isinstance(node, blue.MoveableThingType):
			alpha, beta, gamma = node.alpha, node.beta, node.gamma
		else:
			alpha, beta, gamma = 0, 0, 0
		return cls.E_rot(alpha, beta, gamma)


	@staticmethod
	def X_rot(alpha) -> np.ndarray:
		"""
		Constructs the rotation matrix around the ``X``-axis for a given angle.
		
		Parameters
		----------
		alpha : float
			The angle of rotation.
		
		Returns
		-------
		np.ndarray
			The rotation matrix for the given angle.
		"""
		return np.array([[             1,             0,             0],
				 [             0, np.cos(alpha),-np.sin(alpha)],
				 [             0, np.sin(alpha), np.cos(alpha)]])


	@staticmethod
	def Y_rot(beta) -> np.ndarray:
		"""
		Constructs the rotation matrix around the ``Y``-axis for a given angle.
		
		Parameters
		----------
		beta : float
			The angle of rotation.
		
		Returns
		-------
		np.ndarray
			The rotation matrix for the given angle.
		"""
		return np.array([[ np.cos(beta) ,             0, np.sin(beta) ],
				 [             0,             1,             0],
				 [-np.sin(beta) ,             0, np.cos(beta) ]])


	@staticmethod
	def Z_rot(gamma) -> np.ndarray:
		"""
		Constructs the rotation matrix around the ``Z``-axis for a given angle.
		
		Parameters
		----------
		gamma : float
			The angle of rotation.
		
		Returns
		-------
		np.ndarray
			The rotation matrix for the given angle.
		"""
		return np.array([[ np.cos(gamma) ,-np.sin(gamma),             0],
				 [ np.sin(gamma) , np.cos(gamma),             0],
				 [              0,             0,             1]])
	

	@classmethod
	def E_rot(cls, 
		  alpha = None, 
		  beta  = None, 
		  gamma = None) -> np.ndarray:
		"""
		Constructs the rotation matrix for improper euler angles.
		
		Parameters
		----------
		alpha : float
			The angle around which the ``X``-axis is rotated.
		beta : float
			The angle around which the ``Y``-axis is rotated.
		gamma : float
			The angle around which the ``Z``-axis is rotated.
		
		Returns
		-------
		np.ndarray
			The rotation matrix for the improper euler angles.
		"""
		X = cls.X_rot(alpha or 0)
		Y = cls.Y_rot(beta  or 0)
		Z = cls.Z_rot(gamma or 0)
		R = X @ Y @ Z
		return R


	@classmethod
	def Axis_rot(cls, R) -> np.ndarray:
		"""
		Returns
		-------
		np.ndarray
			The axis of rotation for a given rotation matrix or reference frame
		"""
		S = 0.5 * (R - R.T)
		return np.array([S[2, 1], S[0, 2], S[1, 0]])


	@classmethod
	def quat_to_euler(cls, R, I, J, K) -> tuple[float]:
		"""
		Parameters
		----------
		R : float
			The real rotation angle component of the quaternion.
		I : float
			The ``X``-axis component of the quaternion.
		J : float
			The ``Y``-axis component of the quaternion.
		K : float
			The ``Z``-axis component of the quaternion.

		Returns
		-------
		tuple[float]
			The improper euler angles for a given quaternion.
		"""
		R = np.array([[1 - 2*(J**2 + K**2),       2*(I*J - R*K),       2*(R*J + I*K)],
			      [      2*(I*J + R*K), 1 - 2*(I**2 + K**2),       2*(J*K - R*I)],
			      [      2*(I*K - R*J),       2*(R*I + J*K), 1 - 2*(I**2 + J**2)]])
		return cls.reference_frame_to_euler(R)


	@classmethod
	def euler_to_quat(cls, alpha, beta, gamma) -> tuple[float]:
		"""
		Parameters
		----------
		alpha : float
			The angle around which the ``X``-axis is rotated.
		beta : float
			The angle around which the ``Y``-axis is rotated.
		gamma : float
			The angle around which the ``Z``-axis is rotated.

		Returns
		-------
		tuple[float]
			The quaternion components for improper euler angles.
		"""
		R     = cls.E_rot(alpha, beta, gamma)
		axis  = cls.Axis_rot(R)
		angle = np.arccos(0.5 * (R[0, 0] + R[1, 1] + R[2, 2] - 1))
		alpha_prime = cls.angle(axis, cls.X)
		beta_prime  = cls.angle(axis, cls.Y)
		gamma_prime = cls.angle(axis, cls.Z)
		R = np.cos(angle/2)
		I = np.sin(angle/2) * np.cos(alpha_prime)
		J = np.sin(angle/2) * np.cos(beta_prime)
		K = np.sin(angle/2) * np.cos(gamma_prime)
		return R, I, J, K



class Vector:
	"""
	This class implements multiple functions to modify vectors and orientations of MoveableThings.
	"""
	@classmethod
	def get_axis(cls, axis: str|list[int|float]|np.ndarray):
		"""
		This method returns a vector that represents the axis which is either given as a str 
		or as a 1D array like object.
		
		>>> Vector.get_axis("x")
		array([1., 0., 0.])
		>>> Vector.get_axis([0., 1., 0.])
		array([0., 1., 0.])
		>>> Vector.get_axis(np.array([0., 0., 1.]))
		array([0., 0., 1.])
		
		
		Parameters
		----------
		axis : str | list[int | float] | np.ndarray
			The axis argument may be either a 1D array like object or a string ('x', 'y', 'z').
		
		Returns
		-------
		np.ndarray
			The specified axis converted to a numpy array.
		
		Raises
		------
		ValueError
		If the axis argument is a string and neither 'x', 'y' or 'z' a ValueError is raise.
		"""
		if isinstance(axis, str):
			axis = axis.strip().lower()
			if axis not in ('x', 'y', 'z'):
				raise ValueError(f"Attribute axis must be 'x', 'y' or 'z', got '{axis}' instead.")
			if axis == 'x':
				return np.array([1., 0., 0.], dtype=np.float32)
			elif axis == 'y':
				return np.array([0., 1., 0.], dtype=np.float32)
			elif axis == 'z':
				return np.array([0., 0., 1.], dtype=np.float32)
		else:
			return np.array(axis, dtype=np.float32)


	@staticmethod
	def equal(a: np.ndarray, b: np.ndarray, sensitivity: int = 8):
		"""
		Compares whether two vectors are equal within a margin of error as given by sensitivity.
		
		>>> a = np.array([3.2, 0., 0.])
		>>> b = np.array([3.8, 0., 0.])
		>>> c = np.array([3.2, 5., 0.])
		>>> Vector.equal(a, b, sensitivity=0)
		True
		>>> Vector.equal(a, b, sensitivity=1)
		False
		>>> Vector.equal(a, c, sensitivity=0)
		False
		>>> Vector.equal(a, c, sensitivity=-1)
		True
		
		Parameters
		----------
		a : np.ndarray
			A vector to be compared.
		b : np.ndarray
			A vector to be compared.
		sensitivity : int, optional
			The floating point decimal cutoff.
		
		Returns
		-------
		bool
			A boolean indicating whether the two vectors are equal within the specified margin of error.
		"""
		return np.all(np.round(a - b, sensitivity) == 0)


	@classmethod
	def global_position(cls, node, pos=None):
		"""
		This method converts the local position of a Thing (typically a node in a kinematic tree)
		to its global position relative to the nodes root.
		
		Parameters
		----------
		node : Thing
			A Thing (typically a node in a kinematic tree).
		pos : None, optional
			An initial position that is assumed to be located relative to the node.
		
		Returns
		-------
		np.ndarray
			The global position relative to root of the given node, or the initial position relative to the node is pos was specified.
		"""
		pos = pos or np.zeros(3)
		while node is not None:
			if hasattr(node, 'pos'):
				pos = node.pos + node.rotation_matrix @ pos
			node = node.parent
		return pos
	

	@staticmethod
	def normalize(vector: np.ndarray) -> np.ndarray:
		"""
		This method normalizes a vector to a unit lenth of 1.
		
		Parameters
		----------
		vector : np.ndarray
			A vector to be normalized.
		
		Returns
		-------
		np.ndarray
			The normalized vector.
		"""
		norm = np.linalg.norm(vector)
		assert norm > 0, 'The vector norm must be positive for normalization!'
		return vector/norm


	@staticmethod
	def distance(a: np.ndarray|blue.MoveableThingType, b: np.ndarray|blue.MoveableThingType) -> float:
		"""
		This method returns the distance between two vectors or MoveableThings.
		
		Parameters
		----------
		a : np.ndarray | blue.MoveableThingType
			A vector or MoveableThings global position for which the distance is computed w.r.t. b.
		b : np.ndarray | blue.MoveableThingType
			A vector or MoveableThings global position for which the distance is computed w.r.t. a.
		
		Returns
		-------
		float
			The distance betweent the two vectors or MoveableThings.
		"""
		if not isinstance(a, blue.MoveableThingType):
			a = a.global_pos
		if not isinstance(b, blue.MoveableThingType):
			b = b.global_pos
		return np.linalg.norm(b - a)



#class Distance:
#
#	"""
#	This class implements methods to determin the distance between Things.
#	
#	Attributes
#	----------
#	BOX : tuple
#	A collection of Things that are equivalent in form but inherit from different classes (BaseGeom and BaseSite).
#	CAPSULE : tuple
#	A collection of Things that are equivalent in form but inherit from different classes (BaseGeom and BaseSite).
#	CYLINDER : tuple
#	A collection of Things that are equivalent in form but inherit from different classes (BaseGeom and BaseSite).
#	ELLIPSOID : tuple
#	A collection of Things that are equivalent in form but inherit from different classes (BaseGeom and BaseSite).
#	PLANE : tuple
#	A collection of Things that are equivalent in form but inherit from different classes (BaseGeom and BaseSite).
#	SPHERE : tuple
#	A collection of Things that are equivalent in form but inherit from different classes (BaseGeom and BaseSite).
#	TYPES : dict
#	A dictionary of all forms, containing the other collections.
#	"""
#	
#	CAPSULE   = (blue.geoms.Capsule,   blue.sites.Capsule)
#	CYLINDER  = (blue.geoms.Cylinder,  blue.sites.Cylinder)
#	SPHERE    = (blue.geoms.Sphere,    blue.sites.Sphere)
#	ELLIPSOID = (blue.geoms.Ellipsoid, blue.sites.Ellipsoid)
#	BOX       = (blue.geoms.Box,       blue.sites.Box)
#	PLANE     = (blue.geoms.Plane,)
#	TYPES     = ('CAPSULE', 'CYLINDER', 'PLANE', 'SPHERE', 'ELLIPSOID', 'BOX')
#	@classmethod
#	def type(cls, thing):
#		"""
#		Returns the form type of an object for a given Thing.
#		
#		Parameters
#		----------
#		thing : Thing
#			A Thing either geom or site.
#		
#		Returns
#		-------
#		tuple
#			The collection of classes that implement the form of the given Thing.
#		
#		Raises
#		------
#		NotImplemented
#		If the Things type is not implemented, NotImplemented is raised.
#		"""
#		for TYPE in cls.TYPES:
#			if isinstance(thing, cls.__getattribute__(cls, TYPE)):
#				return TYPE
#		raise NotImplemented
#
#
#	def __getattr__(self, attr):
#		"""
#		This method is used to make min distance calls symetric.
#		
#		Parameters
#		----------
#		attr : str
#			A string of the methods name, for example "_min_CAPSULE_CYLINDER"
#		
#		Returns
#		-------
#		method
#			The distance method reflected in arguments.
#		
#		Raises
#		------
#		AttributeError
#		If the attribute does not adhere to the distance function naming conventions an AttributeError is raised.
#		"""
#		if attr.startswith('_min_'):
#			type_a, type_b = attr.replace('_min_', '').split('_')
#			attr           = f'_min_{type_b}_{type_a}'
#			method         = self.__getattribute__(attr)
#			def wrapper(self, thing_a, thing_b):
#				return method(thing_b, thing_a)
#			return wrapper
#		raise AttributeError
#
#
#	def center(thing_a, thing_b):
#		"""
#		A method that returns the distance between the centers of two Things.
#		
#		Parameters
#		----------
#		thing_a : ThingType
#			Description
#		thing_b : ThingType
#			Description
#		
#		Returns
#		-------
#		TYPE
#			Description
#		"""
#		return Vector.distance(thing_a.pos, thing_b.pos)
#
#
#	@classmethod
#	def minimum(cls, thing_a, thing_b):
#		"""
#		Calculates the minimum distance betweeen two Things analytically.
#		Currently the min Distance is approximated with an upper bound, precise 
#		distance calculations will be unrolled over time.
#		
#		Parameters
#		----------
#		thing_a : ThingType
#			A Thing for which the minimum distance is calculated w.r.t. thing_b.
#		thing_b : ThingType
#			A Thing for which the minimum distance is calculated w.r.t. thing_a.
#		
#		Returns
#		-------
#		float
#			The (currently approximated) minimum Distance between the two Things.
#		"""
#		#### FAST HACK
#		pos_a  = Vector.global_position(thing_a)
#		pos_b  = Vector.global_position(thing_b)
#		size_a = np.max(thing_a.size)
#		size_b = np.max(thing_b.size)
#		distance = Vector.distance(thing_a, thing_b)
#		return distance - size_a - size_b
#		#### FUTURE
#		assert isinstance(thing_a, (blue.geoms.Geom, blue.sites.Site)) and isinstance(thing_b, (blue.geoms.Geom, blue.sites.Site))
#		type_a, type_b = cls.type(thing_a), cls.type(thing_b)
#		method_name = f'_min_{type_a}_{type_b}'
#		method = getattr(cls, method_name)
#		return method(thing_a, thing_b)
#
#
#	@classmethod
#	def _min_SPHERE_SPHERE(cls, thing_a, thing_b):
#		"""
#		Calculates the distance between two Spheres analytically
#		
#		Parameters
#		----------
#		thing_a : Sphere
#			A Sphere for which the minimum distance is calculated analytically w.r.t. thing_b.
#		thing_b : Sphere
#			A Sphere for which the minimum distance is calculated analytically w.r.t. thing_a.
#		
#		Returns
#		-------
#		float
#			The minimum distance between two Spheres.
#		"""
#		return cls.center(thing_a, thing_b) - thing_a.radius - thing_b.radius



if __name__ == '__bluen__':
	R = np.stack([Rotation.Y, Rotation.Z, Rotation.X], axis=1)
	alpha, beta, gamma = Rotation.reference_frame_to_euler(R)
	print(alpha, beta, gamma)
	S1 = blue.geoms.Sphere(radius=1.5)
	S2 = blue.geoms.Sphere(pos=[3.5, 0, 0])
	print(Distance.minimum(S1, S2))


"""
Tendons
=======

:class:`Tendons <blueprints.tendon.Tendon>` can be used to string together multiple Things. The can provide additional 
structural stability or connect otherwise kinetically distinct Things. In Mujoco, 
Tendon can be bound to Sites, Geoms and Joints according to certain constraints 
limiting which Things can be bound and which Tendon attributes are passed into the 
simulation. Tendons introduce cycles to the kinematic hierarchy and are therefore 
treated specially with the :meth:`Path.bind <blueprints.tendon.Path.bind>` method replacing :meth:`NodeThing.attach <blueprints.thing.node.NodeThing.attach>` within 
a special context manager.

Joint (Fixed) Tendons
---------------------
Joint Tendons are the simplest usecase since they only use a reduced set of 
attributes and do not support spliting paths into pulleys. The attributes allowed 
for Joint Tendons are:

* :attr:`name <blueprints.tendon.Tendon.name>`
* :attr:`limited <blueprints.tendon.Tendon.limited>`
* :attr:`range <blueprints.tendon.Tendon.range>`
* :attr:`frictionloss <blueprints.tendon.Tendon.frictionloss>`
* :attr:`margin <blueprints.tendon.Tendon.margin>`
* :attr:`springlength <blueprints.tendon.Tendon.springlength>`
* :attr:`stiffness <blueprints.tendon.Tendon.stiffness>`
* :attr:`damping <blueprints.tendon.Tendon.damping>`

We first create some example Things.

.. code-block:: py
	:caption: Creating Example Things

	>>> world = blue.World()
	>>> body  = blue.Body(geoms=blue.geoms.Sphere(), # introduces mass to Body
	>>>		      joints=blue.joints.Hinge())
	>>> world.attach(body.shift(x=0, name='A'),
	>>> 		 body.shift(x=3, name='B'),
	>>> 		 body.shift(x=6, name='C'))
	>>> tendon = blue.Tendon()

To bind the Joints to the tendon we first grab them from the world and then bind 
them to the tendon path within the Tendons context. When we bind a Joint to a Tendon, 
a coefficient must be specified weighing the interaction strength between Tendon and 
Joint.

.. code-block:: py
	:caption: Tendon Context

	>>> jA = world.bodies['A'].joints[0]
	>>> jB = world.bodies['B'].joints[0]
	>>> jC = world.bodies['C'].joints[0]
	>>> with tendon as path:
	>>> 	 path.bind(jA, coef=1.)
	>>> 	 path.bind(jB, coef=2.)
	>>> 	 path.bind(jC, coef=1.)

.. code-block:: py
	:caption: Alternatively

	>>> with tendon as path:
	>>> 	 path.bind(jA, jB, jC, coef=[1.0, 2.0, 1.0])

The resulting XML is split into the normal kinematic tree and the Tendon referencing 
its bound Things via name attribute.

.. code-block:: mxml
	:caption: XML Structure Tendon
	
	<tendon>
		<fixed name="anonymous_tendon">
			<joint joint="anonymous_hinge_(0)" coef="1.0" />
			<joint joint="anonymous_hinge_(1)" coef="2.0" />
			<joint joint="anonymous_hinge_(2)" coef="1.0" />
		</fixed>
	</tendon>


.. note::
	Keep in mind that Fixed Joint Tendons do not support :attr:`width <blueprints.tendon.Tendon.width>` and are hence invisible.


.. code-block:: mxml
	:caption: XML Structure World

	<worldbody>
		<body name="A">
			<joint type="hinge" name="anonymous_hinge_(0)" axis="0.0 0.0 1.0" />
			<geom size="1.0" type="sphere" name="anonymous_sphere_(0)" />
		</body>
		<body pos="3.0 0.0 0.0" name="B">
			<joint type="hinge" name="anonymous_hinge_(1)" axis="0.0 0.0 1.0" />
			<geom size="1.0" type="sphere" name="anonymous_sphere_(1)"/>
		</body>
		<body pos="6.0 0.0 0.0" name="C">
			<joint type="hinge" name="anonymous_hinge_(2)" axis="0.0 0.0 1.0" />
			<geom size="1.0" type="sphere" name="anonymous_sphere_(2)"/>
		</body>
	</worldbody>


Site/Geom (Spatial) Tendons
---------------------------
Spatial Tendons are bind to :class:`Sites <blueprints.sites.BaseSite>` and :class:`Geoms <blueprints.geoms.BaseGeom>` and support path splitting. The following rules
must be satisfied for mujoco to accept the build.

* Every path must start and end with a :class:`Site <blueprints.sites.BaseSite>`.
* Every :class:`Geom <blueprints.geoms.BaseGeom>` must be sandwiched between two :class:`Sites <blueprints.sites.BaseSite>`.
* Every Tendon binding Sites/Geoms must not also bind Joints.

Lets build some example Things and bind them to a Tendon.


.. code-block:: py
	:caption: Creating Example Things

	>>> body_A = blue.Body(sites=blue.sites.Sphere(radius=0.5)).shift(x=-3, name='A')
	>>> body_B = blue.Body(geoms=blue.geoms.Sphere(radius=0.5),
	>>> 		       sites=blue.sites.Sphere(radius=0.5).shift(z=3), name='B')
	>>> body_C = blue.Body(sites=blue.sites.Sphere(radius=0.5)).shift(x=3, name='C')
	>>> world.attach(body_A, body_C, body_B)
	>>> tendon = blue.Tendon(width=0.1)

When binding a :class:`Geom <blueprints.geoms.BaseGeom>` to a Tendon, an optional ``side_site`` argument may be 
specified. If the Tendon runs through the Geom at the begining of the Mujoco simulation, 
it is snapped out on the side of the specified ``side_site``.

.. code-block:: py
	:caption: Binding Sites and Geoms

	>>> sA = world.bodies['A'].sites[0]
	>>> sB = world.bodies['B'].sites[0]
	>>> gB = world.bodies['B'].geoms[0]
	>>> sC = world.bodies['C'].sites[0]
	>>> with tendon as path:
	>>> 	 path.bind(sA)
	>>> 	 path.bind(gB, side_site=sB)
	>>> 	 path.bind(sC)

.. image:: /_static/tendon_side_site.png

The resulting XML structure are separated into the cyclical references in the header of the Mujoco XML 
and its components in the normal kinematic hierarchy.

.. code-block:: mxml
	:caption: XML Structure Tendon

	<tendon>
		<spatial width="0.1" name="anonymous_tendon">
			<site site="anonymous_sphere_(0)" />
			<geom geom="anonymous_sphere" sidesite="anonymous_sphere_(2)" />
			<site site="anonymous_sphere_(1)" />
		</spatial>
	</tendon>

.. code-block:: mxml
	:caption: XML Structure World

	<worldbody>
		<body pos="-3.0 0.0 0.0" name="A">
		        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(0)" />
		</body>
		<body pos="3.0 0.0 0.0" name="C">
		        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(1)" />
		</body>
		<body name="B">
		        <geom size="0.5" type="sphere" name="anonymous_sphere" />
		        <site size="0.5 0.0 0.0" type="sphere" pos="0.0 0.0 3.0" name="anonymous_sphere_(2)" />
		</body>
	</worldbody>

Path Splitting
--------------
Mujoco supports pullies which split a path into multiple subpaths. The forces acting on each subpath 
are distributed proportionally. The begining of a new path is not necessarily directly connected to 
end of the parent path.

.. code-block:: py
	:caption: Creating Example Things

	>>> world   = blue.World()
	>>> body_A  = blue.Body(pos=[0, 0, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> body_B  = blue.Body(pos=[2, 0, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> body_C1 = blue.Body(pos=[4,-2, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> body_C2 = blue.Body(pos=[4, 2, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> body_D1 = blue.Body(pos=[6,-2, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> body_D2 = blue.Body(pos=[6, 0, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> body_D3 = blue.Body(pos=[6, 4, 0], sites=blue.sites.Sphere(radius=0.5))
	>>> tendon  = blue.Tendon(width=0.1)

Using the method :meth:`split <blueprints.tendon.Path.split>` a pulley is introduced that splits the path into `n` 
branches. A branch path can also be split.

.. code-block:: py
	:caption: Binding with Multiple Paths

	>>> with tendon as path:
	>>> 	 path.bind(body_A.sites)
	>>> 	 path.bind(body_B.sites)
	>>> 	 path_1, path_2, path_3 = path.split(3)
	>>> 	 # PATH ONE
	>>> 	 path_1.bind(body_B.sites)
	>>> 	 path_1.bind(body_C1.sites)
	>>> 	 path_1.bind(body_D1.sites)
	>>> 	 # PATH TWO
	>>> 	 path_2.bind(body_C2.sites)
	>>> 	 path_2.bind(body_D2.sites)
	>>> 	 # PATH THREE
	>>> 	 path_3.bind(body_C2.sites)
	>>> 	 path_3.bind(body_D3.sites)

The example above introduces a pulley with the ``divisor`` 3. The path bindings can be seen in the 
following image from left to right with ``path_1`` on the bottom and ``path_2`` and ``path_3`` ontop both 
sharing the same initial Site.

.. image:: /_static/tendon_example.png

The resulting XML structures are:

.. code-block:: mxml
	:caption: XML Structure Tendon

	<tendon>
                <spatial width="0.1" name="anonymous_tendon">
                        <site site="anonymous_sphere_(0)" />
                        <site site="anonymous_sphere_(1)" />
                        <pulley divisor="3" />
                        <site site="anonymous_sphere_(3)" />
                        <site site="anonymous_sphere_(6)" />
                        <pulley divisor="3" />
                        <site site="anonymous_sphere_(3)" />
                        <site site="anonymous_sphere_(5)" />
                        <pulley divisor="3" />
                        <site site="anonymous_sphere_(1)" />
                        <site site="anonymous_sphere_(2)" />
                        <site site="anonymous_sphere_(4)" />
                </spatial>
        </tendon>


.. code-block:: mxml
	:caption: XML Structure World

	<worldbody>
                <body name="anonymous_body_(0)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(0)" />
                </body>
                <body pos="2.0 0.0 0.0" name="anonymous_body_(1)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(1)" />
                </body>
                <body pos="4.0 -2.0 0.0" name="anonymous_body_(2)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(2)" />
                </body>
                <body pos="4.0 2.0 0.0" name="anonymous_body_(3)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(3)" />
                </body>
                <body pos="6.0 -2.0 0.0" name="anonymous_body_(4)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(4)" />
                </body>
                <body pos="6.0 0.0 0.0" name="anonymous_body_(5)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(5)" />
                </body>
                <body pos="6.0 4.0 0.0" name="anonymous_body_(6)">
                        <site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(6)" />
                </body>
        </worldbody>

Copies of Tendons
-----------------

Copying Tendons takes special attention since Tendons introduce cyclical references (i.e. multiple children of a :class:`NodeThing <blueprints.node.NodeThing>`
can be bound to the same Tendon). Therefore Tendons only get copied when the (root) Thing copied by the user has as its descendants all references 
of the Tendon.

Let's see an example of this:

.. code-block:: py
	:caption: Creating and Binding Example Things

	>>> tendon = blue.Tendon(width=0.1)
	>>> root = blue.Body(name='root')
	>>> body_site = blue.Body(sites=blue.sites.Sphere(radius=0.5))
	>>> root.attach(body_site.shift(x=0, name='A'), 
	>>> 		body_site.shift(x=2, name='B'))
	>>> A = root.bodies['A']
	>>> B = root.bodies['B']
	>>> C = body_site.shift(x=4, name='C')
	>>> with tendon as path:
	>>> 	 path.bind(A.sites)
	>>> 	 path.bind(B.sites)
	>>> 	 path.bind(C.sites)

We have created three bodies with sites (``A``, ``B``, ``C``) that are bound together with a Tendon. 
But only ``A`` and ``B`` are attached to root, while ``C`` is not. If both are attached to the World 
without creating a copy, they world will contain the tendon. 

.. code-block:: py
	:caption: World Attachment without Copy

	>>> world.attach(root, C, copy=False)

.. code-block:: mxml
	:caption: XML Structure Tendon

        <tendon>
                <spatial width="0.1" name="anonymous_tendon">
                        <site site="anonymous_sphere_(0)" />
                        <site site="anonymous_sphere_(1)" />
                        <site site="anonymous_sphere_(2)" />
                </spatial>
        </tendon>

.. code-block:: mxml
	:caption: XML Structure World

	
	<worldbody>
		<body name="root">
			<body name="A">
				<site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(0)" />
			</body>
			<body pos="2.0 0.0 0.0" name="B">
				<site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(1)" />
			</body>
		</body>
		<body pos="4.0 0.0 0.0" name="C">
			<site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(2)" />
		</body>
	</worldbody>


If we however create a copy of ``root`` and ``C`` separately the Tendon will be discarded, since neither 
``root`` nor ``C`` contain all refferences of the Tendon as their descendants (the references of the Tendon 
are the bound Sites ``"anonymous_sphere_(0)"``, ``"anonymous_sphere_(1)"`` and ``"anonymous_sphere_(2)"``).


.. code-block:: py
	:caption: World Attachment without Copy

	>>> world.attach(root, C, copy=True)

.. code-block:: mxml
	:caption: XML Structure Tendon

         </tendon>

.. code-block:: mxml
	:caption: XML Structure World

	<worldbody>
		<body name="root">
			<body name="A">
				<site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(0)" />
			</body>
			<body pos="2.0 0.0 0.0" name="B">
				<site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(1)" />
			</body>
		</body>
		<body pos="4.0 0.0 0.0" name="C">
			<site size="0.5 0.0 0.0" type="sphere" name="anonymous_sphere_(2)" />
		</body>
	</worldbody>
"""



from collections import defaultdict
import numpy as np
import xml.etree.ElementTree as xml
import blueprints as blue



class Tendon(blue.TendonType, blue.NodeThing, blue.ColoredThing, blue.FocalThing):
	@blue.restrict
	def __init__(self, 
		     name:	        str|None    = None, 
		     limited:	        bool|None   = None, 
		     act_force_limited: bool|None   = None, 
		     min_length:	int|float   = 0., 
		     max_length:	int|float   = 0., 
		     min_act_force:     int|float   = 0., 
		     max_act_force:     int|float   = 0., 
		     frictionloss:      int|float   = 0., 
		     width:	        int|float   = 0.003, 
		     color:	        object|None = None, 
		     stiffness:	        int|float   = 0., 
		     damping:	        int|float   = 0., 
		     armature:	        int|float   = 0.,
		     **kwargs) -> None:
		"""
		Some attribute descriptions are taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#tendon>`__.

		Parameters
		----------
		limited : bool | None
			Sets whether length limits constraints are imposed on the mujoco 
			solver. Leaving this argument as None (recommended) enables autolimits 
			derived from the other attributes.
		act_force_limited : bool | None
			This argument is the equivalent to `limited` for actuator forces.
		min_length : int | float
			Minimal length of Tendon
		max_length : int | float
			Maximal length of Tendon
		min_act_force : int | float
			Minimal force allowed to act on Tendon
		max_act_force : int | float
			Maximal force allowed to act on Tendon
		frictionloss : int | float
			Friction loss caused by dry friction. This value should be positive.
		width : int | float
			Crosssection radius of the Tendon
		stiffness : int | float
			If set to a positive value it acts as a Spring coefficient.
		damping : int | float
			If set to a positive value it acts as a Damping coefficient.
		armature : int | float
			Inertia associated with changes in tendon length
		"""
		self._built      = False
		self._ACTIVE	    = False
		# MIGRATION ATTRIBUTES
		self._ADDRESS_BOOK  = defaultdict(list)
		self._MIGRATED      = False
		self._MIGRATIONS    = defaultdict(int)
		# PATH ATTRIBUTES
		self._branches = []
		# CHILDREN
		pack = lambda x: x if isinstance(x, list) else [x]
		self._sites      = []
		self._geoms      = []
		self._side_sites = []
		self._joints     = []
		self._CHILDREN	 = {}
		self._PSEUDO_CHILDREN  = {'sites':       {'type':     blue.SiteType, 
						 	  'children': self._sites}, 
					  'geoms':       {'type':     blue.GeomType, 
							  'children': self._geoms}, 
					  'side_sites':  {'type':     blue.SiteType, 
							  'children': self._side_sites},
					  'joints':      {'type':     blue.JointType, 
							  'children': self._joints}}
		# MUJOCO ATTRIBUTES
		self.limited	   = limited
		self.act_force_limited = act_force_limited
		self.min_length	= min_length
		self.max_length	= max_length
		self.min_act_force     = min_act_force
		self.max_act_force     = max_act_force
		self.frictionloss      = frictionloss
		self.width	     = width
		self.stiffness	 = stiffness
		self.damping	   = damping
		self.armature	  = armature
		# SETTING COLOR BLUEPRINT COLOR DEFAULT FOR GEOMS
		color = color if color is not None else 'grey'
		super().__init__(name=name, 
				 color=color,
				 **kwargs)


	def __enter__(self):
		self._ACTIVE = True
		self._branches = [Path(self)]
		return self._branches[0]


	def __exit__(self, exc_type, error, trace):
		self._ACTIVE = False


	@property
	def _MIGRATION_DONE(self):
		return all(v == 0 for v in self._MIGRATIONS.values())


	def _get_path(self, address):
		path = self
		for i in address:
			path = path._branches[i]
		return path


	def _migrate(self, old, new):
		if not self._MIGRATED:
			self._start_migration()
		if self._VALID:
			addresses = self._ADDRESS_BOOK[old]
			for address, i_path, other in addresses:
				path = self._COPY._get_path(address)
				if isinstance(old, blue.SiteType):
					path._path[i_path][other] = new
					if other == 0:
						self._COPY._sites.append(new)
						new._tendons.append(self._COPY)
					else:
						self._COPY._side_sites.append(new)
						new._side_tendons.append(self._COPY)
				elif isinstance(old, blue.GeomType):
					path._path[i_path][other] = new
					self._COPY._geoms.append(new)
					new._tendons.append(self._COPY)
				elif isinstance(old, blue.JointType):
					path._path[i_path][0] = new
					path._path[i_path][1] = other
					self._COPY._joints.append(new)
					new._tendons.append(self._COPY)

		self._MIGRATIONS[old] -= 1
		if self._MIGRATION_DONE:
			self._finalize_migration()


	def _start_migration(self):
		# CHECK LEGALITY
		self._MIGRATED = True
		self._VALID    = True
		things  = self._sites + self._geoms + self._side_sites + self._joints
		INVALID = []
		for thing in things:
			if blue.REGISTER.copy_root in thing.path:
				self._MIGRATIONS[thing] += 1
			else:
				self._VALID = False
				INVALID.append(thing)
		# CREATE COPY
		self._COPY = super().copy()
		# CREATE MIGRATION OBJECTS
		self._ADDRESS_BOOK = defaultdict(list)
		# CHECK VALIDITY
		if not self._VALID:
			print(f'WARNING: {repr(self)} is discarded. The Things ({', '.join(map(repr, INVALID))}) bound by the Tendon are not descendants of the copied Thing {repr(blue.REGISTER.copy_root)}.')
			return
		# BREADTH SEARCH
		queue = [([i], branch) for i, branch in enumerate(self._branches)]
		while queue:
			address, path = queue.pop(0)
			path_parent = self._COPY._get_path(address[:-1])
			path_copy = Path(self._COPY)
			path_parent._branches.append(path_copy)
			for i_path, (thing, other) in enumerate(path):
				self._ADDRESS_BOOK[thing].append((address, i_path, 0))
				if other is not None and isinstance(other, blue.SiteType):
					self._ADDRESS_BOOK[other].append((address, i_path, 1))
				path_copy._path.append([None, None])
			for i_branch, branch in enumerate(path._branches):
				queue.append((address + [i_branch], branch))
	

	def _finalize_migration(self):
		self._MIGRATED     = False
		self._VALID	= True
		self._MIGRATIONS   = defaultdict(int)
		del self._COPY
		del self._ADDRESS_BOOK


	def attach(self, *items, copy=False) -> None:
		"""
		Attachements are not allowed for Tendons. Use :meth:`Path.bind` instead.
		"""
		raise Exception('Attachement is not allowed for Tendons, use Path.bind instead. Refer to the Documentation for details.')


	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		if self._built:
			return
		mujoco_specs = self._mujoco_specs(**kwargs)
		if self._joints:
			INVALID = []
			for attr in mujoco_specs.keys():
				if attr not in self._FIXED_ATTR:
					INVALID.append(attr)
			for attr in INVALID:
				del mujoco_specs[attr]
			if INVALID:
				print(f'WARNING: Tendons with Joints have a reduced set of attributes. Ignoring attributes ({', '.join(map(str, INVALID))}).')
		tendon = xml.SubElement(world._xml_tendon, 
					'fixed' if self._joints else 'spatial', 
					**mujoco_specs)
		queue = [(len(self._branches), path) for path in self._branches]
		while queue:
			idx, path = queue.pop()
			if idx > 1:
				pulley = xml.SubElement(tendon, 'pulley', divisor=str(idx))
			for thing, other in path:
				if isinstance(thing, blue.SiteType):
					xml.SubElement(tendon, 'site', site=thing.name)
				elif isinstance(thing, blue.GeomType):
					if not isinstance(other, blue.SiteType):
						xml.SubElement(tendon, 'geom', geom=thing.name)
					else:
						xml.SubElement(tendon, 'geom', geom=thing.name, sidesite=other.name)
				elif isinstance(thing, blue.JointType):
					if not isinstance(other, (int, float)):
						xml.SubElement(tendon, 'joint', joint=thing.name)
					else:
						xml.SubElement(tendon, 'joint', joint=thing.name, coef=str(float(other)))

			for branch in path._branches[::-1]:
				queue.insert(0, (len(path._branches), branch))
		self._built = True

	# DERIVED ATTRIBUTES

	@property
	def actuatorfrclimited(self) -> bool|None:
		"""
		Indicates whether the Actuator forces applied are limited.

		Returns
		-------
		bool | None
			In the attribute is None, the value is set to ``auto``.
		"""
		return self._act_force_limited


	@property
	def range(self) -> np.ndarray:
		"""
		The length range of the Tendon.

		Returns
		-------
		np.ndarray
			The two entries indicate minimum and maximum Tendon length.
		"""
		return np.array([self.min_length, 
				 self.max_length], dtype=np.float32)


	@property
	def actuatorfrcrange(self) -> np.ndarray:
		"""
		The Actuator force range of the Tendon.

		Returns
		-------
		np.ndarray
			The two entries indicate minimum and maximum Actuator force applied to the Tendon.
		"""
		return np.array([self.min_act_force, 
				 self.max_act_force], dtype=np.float32)

	# BLUEPRINTS ATTRIBUTES

	@property
	def max_length(self) -> float:
		"""
		Changing this value results in a change in the first component of ``range``.

		The user should make sure, that the initial length of the tendon as 
		computed by the distance between its bound Things does not exeed this attribute.

		Returns
		-------
		float
			Maximum length of the Tendon
		"""
		return self._max_length


	@max_length.setter
	@blue.restrict
	def max_length(self, max_length: int|float) -> None:
		self._max_length = float(max_length)


	@property
	def min_length(self) -> float:
		"""
		Changing this value results in a change in the second component of ``range``.
		
		Returns
		-------
		float
			Minimum length of the Tendon
		"""
		return self._min_length


	@min_length.setter
	@blue.restrict
	def min_length(self, min_length: int|float) -> None:
		self._min_length = float(min_length)


	@property
	def max_act_force(self) -> float:
		"""
		Changing this value results in a change in the first component of ``actuatorfrcrange``.

		Returns
		-------
		float
			Maximum force applicable by an :class:`Actuator <blueprints.actuator.BaseActuator>`
		"""
		return self._max_act_force


	@max_act_force.setter
	@blue.restrict
	def max_act_force(self, max_act_force: int|float) -> None:
		self._max_act_force = float(max_act_force)


	@property
	def min_act_force(self) -> float:
		"""
		Changing this value results in a change in the second component of ``actuatorfrcrange``.

		Returns
		-------
		float
			Minimum force applicable by an :class:`Actuator <blueprints.actuator.BaseActuator>`
		"""
		return self._min_act_force


	@min_act_force.setter
	@blue.restrict
	def min_act_force(self, min_act_force: int|float) -> None:
		self._min_act_force = float(min_act_force)

	# MUJOCO ATTRIBUTES

	@property
	def limited(self) -> bool|None:
		"""
		Indicates whether the Tendon length is limited by ``range``.

		Returns
		-------
		bool | None
			In the attribute is None, the value is set to ``auto``.
		"""
		return self._limited


	@limited.setter
	@blue.restrict
	def limited(self, limited: bool|None) -> None:
		self._limited = limited


	@property
	def act_force_limited(self) -> bool|None:
		"""
		Indicates whether the Actuator forces applied are limited.

		Returns
		-------
		bool | None
			In the attribute is None, the value is set to ``auto``.
		"""
		return self._act_force_limited


	@act_force_limited.setter
	@blue.restrict
	def act_force_limited(self, act_force_limited: bool|None) -> None:
		self._act_force_limited = act_force_limited


	@property
	def frictionloss(self) -> float:
		"""
		Friction loss caused by dry friction.

		Returns
		-------
		float
			To enable friction loss, set this attribute to a positive value.
		"""
		return self._frictionloss


	@frictionloss.setter
	@blue.restrict
	def frictionloss(self, frictionloss: int|float) -> None:
		self._frictionloss = float(frictionloss)


	@property
	def width(self) -> float:
		"""
		This attribute is purely cosmetical.

		Returns
		-------
		föoat
			Width of the Tendon.
		"""
		return self._width


	@width.setter
	@blue.restrict
	def width(self, width: int|float) -> None:
		self._width = float(width)


	@property
	def stiffness(self) -> float:
		"""
		A positive value generates a spring force (linear in position) acting along the Tendon.

		Returns
		-------
		float
			Stiffness of the Tendon
		"""
		return self._stiffness


	@stiffness.setter
	@blue.restrict
	def stiffness(self, stiffness: int|float) -> None:
		self._stiffness = float(stiffness)


	@property
	def damping(self) -> float:
		"""
		A positive value generates a damping force (linear in velocity) acting along the tendon. 
		If possible, :attr:`Joint.damping <blueprints.joints.BaseJoint.damping>` should be used 
		for stability.

		Returns
		-------
		float
			Tendon dampin.
		"""
		return self._damping


	@damping.setter
	@blue.restrict
	def damping(self, damping: int|float) -> None:
		self._damping = float(damping)


	@property
	def armature(self) -> float:
		"""
		Setting this attribute to a positive value :math:`m` adds a kinetic 
		energy term :math:`\\frac{1}{2}mv^2`, where :math:`v` is the tendon velocity.

		Returns
		-------
		float
			Inertia associated with changes in Tendon length
		"""
		return self._armature


	@armature.setter
	@blue.restrict
	def armature(self, armature: int|float) -> None:
		self._armature = float(armature)



class Path(blue.PathType):
	"""
	Paths are used to specify the ``<pulley>`` element in Mujoco. 
	A path can be strung along multiple :class:`Sites <blueprints.sites.BaseSite>`, 
	:class:`Joints <blueprints.joints.BaseJoint>` and :class:`Geoms <blueprints.geoms.BaseGeom>`. 
	Every path must start and end with a Site and every Geom must be sandwitched between two Sites.
	"""
	def __init__(self, tendon):
		"""
		.. warning::
			This class is not to be instantiated by the user directly. Use the contextmanager 
			:meth:`Tendon.__enter__` instead (see above).
		"""
		self.tendon = tendon
		self._path  = []
		self._branches = []
		self._split = False


	def __iter__(self):
		for thing, side_site in self._path:
			yield thing, side_site


	def __getitem__(self, idx):
		if isinstance(idx, int):
			if idx >= len(self._path):
				raise IndexError(f'This Path has only {len(self._path)} entries but entry {idx} was requested!')
			return self._path[idx][0]
		elif isinstance(idx, slice):
			if idx.start >= len(self._path):
				raise IndexError(f'This Path has only {len(self._path)} entries but entry {idx.start} was requested!')
			elif idx.stop >= len(self._path):
				raise IndexError(f'This Path has only {len(self._path)} entries but entry {idx.stop} was requested!')
			return [thing for thing, side_site in self._path[idx]]
		else:
			raise NotImplemented


	@blue.restrict
	def bind(self, 
		 *things:   list[blue.ThingType], 
		 side_site: blue.SiteType|None = None,
		 coef:      int|float|list[int|float]|None = None) -> None:
		"""
		Binding is performed in order. If a ``side_site`` is specified only 
		one Geom to be bound can be passed.

		Parameters
		----------
		*things : list [ blue.ThingType ]
			The Sites and Geoms to be bound to a Tendon path
		side_site: blue.SiteType | None
			A side Site for a Geom specifies to which side of the geom 
			the Tendon should snap out if passing through the Geom initially.
		coef : int | float | list [ int | float] | None
			Numeric coefficients for Joints. The argument must have the same shape as *things.
		"""
		if self._split:
			raise Exception(f'Paths can not bind Things after they were split. ')
		if not self.tendon._ACTIVE:
			raise Exception('Tendons can only be bound to Things within the context manager. Use "with tendon as path:" and refer to the Documentation.')
		if side_site is not None:
			if len(things) != 1 or not isinstance(things[0], blue.GeomType):
				raise ValueError('If a side_site is specified, precisely one Geom must be given.')
		# FLATTENING VIEWS
		flattened_things = []
		for thing in things:
			if isinstance(thing, blue.ViewType):
				flattened_things.extend(list(thing))
			else:
				flattened_things.append(thing)
		# CHECKING COEFFICIENTS
		if coef is not None:
			if not all(isinstance(thing, blue.JointType) for thing in flattened_things):
				raise ValueError(f'Coefficients can only be specified for Joints.')
			coef = [coef] if not isinstance(coef, list) else coef
			if len(coef) != len(things):
				raise ValueError(f'Every Coefficient must correspond to a Joint. Got {len(coef)} Coefficients and {len(things)} Joints.')
		# BINDING THINGS
		for i, thing in enumerate(flattened_things):
			if isinstance(thing, blue.SiteType):
				if self.tendon._joints:
					raise ValueError('Tendons can either bind Joints or Sites and Geoms, not both!')
				self.tendon._sites.append(thing)
			elif isinstance(thing, blue.GeomType):
				if isinstance(self._path[-1], blue.GeomType):
					raise ValueError('Geoms bound to a Tendon must be sandwiched by Sites!')
				if self.tendon._joints:
					raise ValueError('Tendons can either bind Joints or Sites and Geoms, not both!')
				if not isinstance(thing, (blue.SphereGeomType, blue.CylinderGeomType)):
					raise ValueError(f'The only geom types allowed for Tendon binding are Cylinders and Spheres. Got {type(thing)}.')
				self.tendon._geoms.append(thing)
				if side_site is not None:
					self.tendon._side_sites.append(side_site)
					side_site._side_tendons.append(self.tendon)
			elif isinstance(thing, blue.JointType):
				if self.tendon._geoms or self.tendon._sites:
					raise ValueError('Tendons can either bind Joints or Sites and Geoms, not both!')
				self.tendon._joints.append(thing)
			else:
				raise ValueError(f'Tendon.bind expects Things of Type Geom, Site or Joint. Got {', '.join(map(lambda x: str(repr(x)), flattened_things))}')
			other = coef[i] if coef is not None and isinstance(thing, blue.JointType) else side_site
			self._path.append([thing, other])
			thing._tendons.append(self.tendon)


	def split(self, number: int) -> tuple:
		"""
		If a pulley should split the Tendon into :math:`n` :class:`Paths <blueprints.tendon.Path>`, this method 
		handles the split. A split Path does not necessarily start at the last element of its parent Path. 
		Once split, a Path can no longer bind Things.

		Parameters
		----------
		number : int
			The number of new Paths

		Returns
		-------
		tuple [ blue.PathType ]
			The new Paths connected through a pulley
		"""
		if not self.tendon._ACTIVE:
			raise Exception('Paths can only be split within the context manager. Use "with tendon as path:" and refer to the Documentation.')
		if self.tendon._joints:
			raise ValueError('Tendons binding Joints cannot be split into pulleys!')
		self._branches = list(Path(self.tendon) for _ in range(number))
		self._split = True
		return tuple(iter(self._branches))


import blueprints as blue
import xml.etree.ElementTree as xml
import numpy as np
import mujoco



class BaseSensor(blue.SensorType, blue.thing.BaseThing):

	"""
	Sensors can be used to as an interface for the agent to the World. Sensors that are part of the 
	kinematic tree of an Agent are included in the Agents observations.
	
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#sensor>`__.
	"""

	@blue.restrict
	def __init__(self, 
		     noise:     float|int           = 0., 
		     cutoff:    float|int           = 0., 
		     name:      str|None            = None, 
		     reference: blue.ThingType|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		noise : float | int, optional
		   	The standard deviation of zero-mean Gaussian noise added to the sensor output.
		cutoff : float | int, optional
			When this value is positive, it limits the absolute value of the sensor output.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict 
			by appending an enumeration scheme.
		reference : blue.ThingType | None, optional
			The reference of the Sensor attribute is the parent of the Sensor which is translated 
			to the :attr:`site <SiteSensor.site>`, :attr:`joint <JointSensor.joint>` or 
			:attr:`actuator <ActuatorSensor.actuator>` attributes for xml.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		
		Raises
		------
		TypeError
			If the reference type is not valid an error is raised.
		"""
		if not isinstance(reference, tuple(self._REFERENCE_TYPES)) and reference is not None:
			raise TypeError(f'The Sensor reference must be of the types ({", ".join(map(str, self._REFERENCE_TYPES))}), got ({type(reference)}) instead.')
		self.noise  = noise
		self.cutoff = cutoff
		super().__init__(name=name, **kwargs)
		self.reference = reference


	@property
	def observation(self) -> np.ndarray:
		"""
		The live sensor data from the simulation.
		
		Returns
		-------
		np.ndarray
		"""
		if not hasattr(self, '_index'):
			raise Exception('Sensor must first be build by a World before observations are available.')
		else:
			return self.root._mj_data.sensordata[self._index:self._index + self.DIMENSIONS].copy()


	@blue.restrict
	def _build(self, 
		   parent, 
		   world:    blue.WorldType, 
		   indicies: dict, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Sensor.
		"""
		self._index          = indicies['sensors']
		indicies['sensors'] += self.DIMENSIONS
		self._xml_root       = xml.SubElement(world._xml_sensor, 
						      self.type,  
						      **self._mujoco_specs(kwargs))
		return self._xml_root


	@blue.restrict
	@classmethod
	def _from_xml_element(cls, xml_element: xml.Element) -> blue.ThingType:
		"""
		This method reconstructs a Thing from an xml element. If any argument for 
		an inheriting class has to be set manually this method must be overwritten.
		
		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a Thing is reconstructed.
		
		Returns
		-------
		blue.ThingType
			The reconstructed Sensor.
		"""
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		sensor_type = xml_element.tag
		sensor = object.__new__(blue.REGISTER.SENSOR_THINGS[sensor_type])
		sensor.__init__(**init_args)
		for key, val in post_args.items():
			setattr(sensor, key, val)
		return sensor


	@property
	def reference(self):
		"""
		The reference of the Sensor attribute is the parent of the Sensor which is translated 
		to the :attr:`site <SiteSensor.site>`, :attr:`joint <JointSensor.joint>` or 
		:attr:`actuator <ActuatorSensor.actuator>` attributes for xml.

		Returns
		-------
		blue.ThingType
			The sensors refferenced Thing
		"""
		return self.parent


	@reference.setter
	@blue.restrict
	def reference(self, reference: blue.NodeThingType|None):
		"""
		Parameters
		----------
		reference : blue.NodeThingType | None
			The reference of the Sensor attribute is the parent of the Sensor which is translated 
			to the :attr:`site <SiteSensor.site>`, :attr:`joint <JointSensor.joint>` or 
			:attr:`actuator <ActuatorSensor.actuator>` attributes for xml.
		"""
		if self.parent is not None:
			self.parent.detach(self)
		if reference is not None:
			reference.attach(self, copy=False)

	# MUJOCO PROPERTIES

	@property
	def type(self):
		"""
		This is a derived attribute, which is used to specify the ``type`` attribute in the 
		mujoco ``sensor`` tag.

		Returns
		-------
		str
			The type is the lower case name of the Sensor class.
		"""
		return self.__class__.__name__.lower()


	@property
	def noise(self) -> float:
		"""
		The standard deviation of zero-mean Gaussian noise added to the sensor output.
		
		Returns
		-------
		float
		"""
		return self._noise


	@noise.setter
	@blue.restrict
	def noise(self, noise: float|int) -> None:
		"""
		Parameters
		----------
		noise : float | int
			The standard deviation of zero-mean Gaussian noise added to the sensor output.
		"""
		self._noise = float(noise)


	@property
	def cutoff(self) -> float:
		"""
		When this value is positive, it limits the absolute value of the sensor output.
		
		Returns
		-------
		float
		"""
		return self._cutoff


	@cutoff.setter
	@blue.restrict
	def cutoff(self, cutoff: float|int) -> None:
		"""
		Parameters
		----------
		cutoff : float | int
			When this value is positive, it limits the absolute value of the sensor output.
		"""
		self._cutoff = float(cutoff)



# INTERMEDIATE SENSORS



class SiteSensor(blue.SiteSensorType, BaseSensor):

	"""
	Intermediate Sensor class used by Sensors that reference :class:`Sites <blueprints.sites.BaseSite>`.
	"""
	
	@property
	def site(self):
		"""
		The parent name of the Sensor which is translated to this attribute for xml construction.
		
		Returns
		-------
		blue.ThingType
		"""
		if self.reference:
			return self.reference.name
		else:
			return None



class JointSensor(blue.JointSensorType, BaseSensor):

	"""
	Intermediate Sensor class used by Sensors that reference :class:`Joints <blueprints.joints.BaseJoint>`.
	"""
	
	@property
	def joint(self):
		"""
		The parent name of the Sensor which is translated to this attribute for xml construction.
		
		Returns
		-------
		blue.ThingType
		"""
		if self.reference:
			return self.reference.name
		else:
			return None



class ActuatorSensor(blue.ActuatorSensorType, BaseSensor):

	"""
	Intermediate Sensor class used by Sensors that reference :class:`Actuators <blueprints.actuators.BaseActuator>`.
	"""
	
	@property
	def actuator(self):
		"""
		The parent name of the Sensor which is translated to this attribute for xml construction.
		
		Returns
		-------
		blue.ThingType
		"""
		if self.reference:
			return self.reference.name
		else:
			return None



# INFO LASER



class InfoLaser(blue.InfoLaserType, SiteSensor):

	"""
	The InfoLaser is not a mujoco native object but a derived blueprints object, 
	that will not appear in xml. Instead it uses the mujoco.mj_ray subroutine to 
	cast a ray from the Sensors Site and returns the intersected object and the 
	distance to it.
	"""
	
	@blue.restrict
	def __init__(self, 
		     axis:      np.ndarray|list[int|float] = [0., 0., 1.], 
		     noise:     float|int                  = 0., 
		     cutoff:    float|int                  = 0., 
		     name:      str|None                   = None, 
		     reference: blue.ThingType|None        = None, 
		     **kwargs):
		"""
		Parameters
		----------
		axis : np.ndarray
			The local axis along which the InfoLasers ray will be casted.
		noise : float | int, optional
		   	The standard deviation of zero-mean Gaussian noise added to the sensor output.
		cutoff : float | int, optional
			When this value is positive, it limits the absolute value of the sensor output.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict 
			by appending an enumeration scheme.
		reference : blue.ThingType | None, optional
			The reference of the Sensor attribute is the parent of the Sensor which is translated 
			to the :attr:`site <SiteSensor.site>`, :attr:`joint <JointSensor.joint>` or 
			:attr:`actuator <ActuatorSensor.actuator>` attributes for xml.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		
		Raises
		------
		TypeError
			If the reference type is not valid an error is raised.
		"""
		self.axis = axis
		super().__init__(noise=noise, 
				 cutoff=cutoff, 
				 name=name, 
				 reference=reference, 
				 **kwargs)


	@property
	def observation(self) -> dict:
		"""
		Computes the observation on runtime.
		"""
		if not self.root._launched:
			raise Exception('Sensor must first be build by a World before observations are available.')
		else:
			R  = self.parent.global_rotation_matrix
			ID = np.array([-1], np.int32)
			distance = mujoco.mj_ray(m=self.root._mj_model, 
						 d=self.root._mj_data, 
						 pnt=self.parent.global_pos, 
						 vec=R @ self.axis, 
						 geomgroup=None, 
						 flg_static=1, 
						 bodyexclude=-1, 
						 geomid=ID)
			if distance == -1:
				geom = None
			else:
				geom = self.root.descendants['geoms']['descendants'][ID[0]]
			return {'distance': distance, 'geom': geom}



	@property
	def axis(self) -> np.ndarray:
		"""
		The local axis along which the InfoLasers ray will be casted.
		
		Returns
		-------
		np.ndarray
		"""
		return self._axis


	@axis.setter
	@blue.restrict
	def axis(self, 
		 axis: np.ndarray|list[int|float]) -> None:
		"""
		Parameters
		-------
		axis : np.ndarray
			The local axis along which the InfoLasers ray will be casted.
		"""
		self._axis = np.array(axis, dtype=np.float32)


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		InfoLasers are not part of mujoco xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The parent of the Thing
		world : blue.WorldType
			The World from which the initial _meth:`blueprints.world.World.build` method was called.
		"""
		pass


# SITE SENSORS



class Touch(SiteSensor):
	DIMENSIONS = 1



class Accelerometer(SiteSensor):
	DIMENSIONS = 3



class Velocimeter(SiteSensor):
	DIMENSIONS = 3



class Gyro(SiteSensor):
	DIMENSIONS = 3



class Force(SiteSensor):
	DIMENSIONS = 3



class Torque(SiteSensor):
	DIMENSIONS = 3



class Rangefinder(SiteSensor):
	DIMENSIONS = 1



# JOINT SENSORS



class JointPos(JointSensor):
	DIMENSIONS = 1



class JointVel(JointSensor):
	DIMENSIONS = 1



class JointLimitPos(JointSensor):
	DIMENSIONS = 1



class JointLimitVel(JointSensor):
	DIMENSIONS = 1



class JointLimitFrc(JointSensor):
	DIMENSIONS = 1



class BallQuat(JointSensor):
	DIMENSIONS = 4



class BallAngVel(JointSensor):
	DIMENSIONS = 3



# ACTUATOR SENSOR



class ActuatorPos(ActuatorSensor):
	DIMENSIONS = 1



class ActuatorVel(ActuatorSensor):
	DIMENSIONS = 1



class ActuatorFrc(ActuatorSensor):
	DIMENSIONS = 1



SENSOR_THINGS  = {'touch':         Touch, 
		  'accelerometer': Accelerometer, 
		  'velocimeter':   Velocimeter, 
		  'gyro':          Gyro, 
		  'force':         Force, 
		  'torque':        Torque, 
		  'rangefinder':   Rangefinder, 
		  'jointpos':      JointPos, 
		  'jointvel':      JointVel, 
		  'jointlimitpos': JointLimitPos, 
		  'jointlimitvel': JointLimitVel, 
		  'jointlimitfrc': JointLimitFrc, 
		  'ballquat':      BallQuat, 
		  'ballangvel':    BallAngVel, 
		  'actuatorpos':   ActuatorPos, 
		  'actuatorvel':   ActuatorVel, 
		  'actuatorfrc':   ActuatorFrc}
